# -*- coding: utf-8 -*-
"""
Demos and examples (:mod:`qit.examples`)
========================================

.. currentmodule:: qit.examples

This module contains various examples which demonstrate the features
of QIT. The :func:`tour` function is a "guided tour" to the toolkit
which runs all the demos in succession.


Demos
-----

.. autosummary::

   adiabatic_qc_3sat
   bb84
   bernstein_vazirani
   grover_search
   markov_decoherence
   nmr_sequences
   phase_estimation_precision
   qft_circuit
   quantum_channels
   quantum_walk
   qubit_and_resonator
   shor_factorization
   superdense_coding
   teleportation
   tour


General-purpose quantum algorithms
----------------------------------

.. autosummary::

   adiabatic_qc
   phase_estimation
   find_order
"""
# Ville Bergholm 2011-2014

from __future__ import division, absolute_import, print_function, unicode_literals

from math import asin
from operator import mod
from copy import deepcopy

import numpy as np
from numpy import (asarray, array, diag, kron, prod, floor, ceil, sqrt, log2, exp, angle, arange, linspace,
    logical_not, sin, cos, arctan2, empty, zeros, ones, eye, sort, nonzero, pi, trace, dot, meshgrid, r_)
from numpy.linalg import eig, norm
import numpy.random as npr
from scipy.misc import factorial
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D   # FIXME makes buggy 3d plotting work for some reason


from .base import sx, sy, sz, p0, p1, tol
from .lmap import lmap, tensor, numstr_to_array
from .utils import gcd, lcm, qubits, R_y, rand_U, rand_SU, mkron, boson_ladder
from .state import state, fidelity
from . import gate, ho, plot


__all__ = [
 'adiabatic_qc_3sat',
 'adiabatic_qc',
 'bb84',
 'bernstein_vazirani',
 'grover_search',
 'markov_decoherence',
 'nmr_sequences',
 'phase_estimation',
 'phase_estimation_precision',
 'qft_circuit',
 'quantum_channels',
 'quantum_walk',
 'qubit_and_resonator',
 'shor_factorization',
 'find_order',
 'superdense_coding',
 'teleportation',
 'tour']


def adiabatic_qc_3sat(n=6, n_clauses=None, clauses=None, problem='3sat'):
    """Adiabatic quantum computing demo.

    This example solves random 3-SAT problems by simulating the
    adiabatic quantum algorithm of :cite:`Farhi`.

    Note that this is incredibly inefficient because we first essentially
    solve the NP-complete problem classically using an exhaustive
    search when computing the problem Hamiltonian H1, and then
    simulate an adiabatic quantum computer solving the same problem
    using the quantum algorithm.
    """
    # Ville Bergholm 2009-2011

    print('\n\n=== Solving 3-SAT using adiabatic qc ===\n')

    if n < 3:
        n = 3

    if clauses == None:
        if n_clauses == None:
            if problem == '3sat':
                n_clauses = 5*n
            else:  # exact cover
                n_clauses = n//2

        # generate clauses
        clauses = zeros((n_clauses, 3), int)
        for k in range(n_clauses):
            bits = list(range(n))
            for j in range(3):
                clauses[k, j] = bits.pop(npr.randint(len(bits))) + 1 # zero can't be negated, add one
        clauses = sort(clauses, 1)

        if problem == '3sat':
            for k in range(n_clauses):
                for j in range(3):
                    clauses[k, j] *= (-1) ** (npr.rand() < 0.5) # negate if bit should be inverted
    else:
        n_clauses = clauses.shape[0]

    # cache some stuff (all the matrices in this example are diagonal, so)
    zb  = array([0, 1], int) # 0.5*(I - sz)
    z_op = []
    for k in range(n):
        z_op.append(mkron(ones(2 ** k, int), zb, ones(2 ** (n-k-1), int)))

    print('{0} bits, {1} clauses'.format(n, n_clauses))
    # Encode the problem into the final Hamiltonian.
    H1 = 0
    if problem == '3sat':
        b = [0, 0, 0]
        for k in range(n_clauses):
            # h_c = (b1(*) v b2(*) v b3(*))
            for j in range(3):
                temp = clauses[k, j]
                b[j] = z_op[abs(temp) - 1] # into zero-based index
                if temp < 0:
                    b[j] = logical_not(b[j])
            H1 += logical_not(b[0] + b[1] + b[2]) # the not makes this a proper OR
        print('Find a bit string such that all the following "(b_i or b_j or b_k)" clauses are satisfied.')
        print('Minus sign means the bit is negated.')
    else:
        # exact cover
        for k in range(n_clauses):
            # h_c = (b1 ^ b2* ^ b3*) v (b1* ^ b2 ^ b3*) v (b1* ^ b2* ^ b3)
            b1 = z_op[clauses[k, 0] - 1]
            b2 = z_op[clauses[k, 1] - 1]
            b3 = z_op[clauses[k, 2] - 1]
            s1 = b1 * logical_not(b2) * logical_not(b3)
            s2 = b2 * logical_not(b3) * logical_not(b1)
            s3 = b3 * logical_not(b1) * logical_not(b2)
            H1 += logical_not(s1 + s2 + s3 -s1*s2 -s2*s3 -s3*s1 +s1*s2*s3)
        print('Find a bit string such that all the following "exactly one of bits {a, b, c} is 1" clauses are satisfied:')
    print(clauses)

    # build initial Hamiltonian
    xb = 0.5 * (eye(2) -sx)
    H0 = 0
    for k in range(n):
        H0 += mkron(eye(2 ** k), xb, eye(2 ** (n-k-1)))

    # initial state (ground state of H0)
    s0 = state(ones(2 ** n) / sqrt(2 ** n), qubits(n)) # n qubits, uniform superposition

    # adiabatic simulation
    adiabatic_qc(H0, H1, s0)
    return H1, clauses




def adiabatic_qc(H0, H1, s0, tmax=50):
    """Adiabatic quantum computing.

    This is a helper function for simulating the adiabatic quantum
    algorithm of :cite:`Farhi` and plotting the results.
    """
    # Ville Bergholm 2009-2011

    H1_full = diag(H1) # into a full matrix

    # adiabatic simulation
    steps = tmax*10
    t = linspace(0, tmax, steps)

    # linear path
    H_func = lambda t: (1-t/tmax)*H0 +(t/tmax)*H1_full
    res = s0.propagate(H_func, t)

    # plots
    # final state probabilities
    plt.figure()
    plot.adiabatic_evolution(t, res, H_func)

    plt.figure()
    res[-1].plot()
    plt.title('Final state')

    print('Final Hamiltonian (diagonal):')
    print(H1)

    print('Measured result:')
    dummy, dummy, res = res[-1].measure(do = 'C')
    print(res)
    if H1[nonzero(res.data)[0]] == 0:
        print('Which is a valid solution!')
    else:
        print('Which is not a solution!')
        if np.min(H1) > 0:
            print("(In this problem instance there aren't any.)")



def bb84(n=50):
    """Bennett-Brassard 1984 quantum key distribution protocol demo.

    Simulate the protocol with n qubits transferred.

    """
    # Ville Bergholm 2009-2012
    from .base import H

    print('\n\n=== BB84 protocol ===\n')
    print('Using {0} transmitted qubits, intercept-resend attack.\n'.format(n))

    # Alice generates two random bit vectors
    sent    = npr.rand(n) > 0.5
    basis_A = npr.rand(n) > 0.5

    # Bob generates one random bit vector
    basis_B = npr.rand(n) > 0.5
    received = zeros(n, bool)

    # Eve hasn't yet decided her basis
    basis_E = zeros(n, bool)
    eavesdrop = zeros(n, bool)

    print("""Alice transmits a sequence of qubits to Bob using a quantum channel.
For every qubit, she randomly chooses a basis (computational or diagonal)
and randomly prepares the qubit in either the |0> or the |1> state in that basis.""")
    print("\nbasis_A = {0}\nbits_A  = {1}".format(asarray(basis_A, dtype=int),
                                                  asarray(sent, dtype=int)))

    temp = state('0')
    for k in range(n):
        # Alice has a source of qubits in the zero state.
        q = temp

        # Should Alice flip the qubit?
        if sent[k]:  q = q.u_propagate(sx)

        # Should Alice apply a Hadamard?
        if basis_A[k]:  q = q.u_propagate(H)

        # Alice now sends the qubit to Bob...
        # ===============================================
        # ...but Eve intercepts it, and conducts an intercept/resend attack

        # Eve might have different strategies here... TODO
        # Eve's strategy (simplest choice, random)
        basis_E[k] = npr.rand() > 0.5
  
        # Eve's choice of basis: Hadamard?
        if basis_E[k]:  q = q.u_propagate(H)

        # Eve measures in the basis she has chosen
        _, res, q = q.measure(do = 'C')
        eavesdrop[k] = res

        # Eve tries to reverse the changes she made...
        if basis_E[k]:  q = q.u_propagate(H)

        # ...and sends the result to Bob.
        # ===============================================

        # Bob's choice of basis
        if basis_B[k]:  q = q.u_propagate(H)

        # Bob measures in the basis he has chosen, and discards the qubit.
        _, res = q.measure()
        received[k] = res

    #sum(xor(sent, eavesdrop))/n
    #sum(xor(sent, received))/n

    print("""\nHowever, there's an eavesdropper, Eve, on the line. She intercepts the qubits,
randomly measures them in either basis (thus destroying the originals!), and then sends
a new batch of qubits corresponding to her measurements and basis choices to Bob.
Since Eve on the average can choose the right basis only 50% of the time,
about 1/4 of her bits differ from Alice's.""")
    print("\nbasis_E = {0}\nbits_E  = {1}".format(asarray(basis_E, dtype=int),
                                                  asarray(eavesdrop, dtype=int)))

    print("\nWhen Bob receives the qubits, he randomly measures them in either basis.")
    print("\nbasis_B = {0}\nbits_B  = {1}".format(asarray(basis_B, dtype=int),
                                                  asarray(received, dtype=int)))

    print("""\nNow Bob announces on a public classical channel that he has received all the qubits.
Alice and Bob then reveal the bases they used. Whenever the bases happen to match,
(about 50% of the time on the average), they both add their corresponding bit to
their personal key. The two keys should be identical unless there's been an eavesdropper.""")

    match = np.logical_not(np.logical_xor(basis_A, basis_B))
    key_A = sent[match]
    key_B = received[match]
    m = len(key_A)
    print('\nmatch = {0}\nkey_A = {1}\nkey_B = {2}'.format(asarray(match, dtype=int),
                                                           asarray(key_A, dtype=int),
                                                           asarray(key_B, dtype=int)))
    print('\nMismatch frequency between Alice and Bob: {0}'.format(np.sum(np.logical_xor(key_A, key_B)) / m))

    print("""\nAlice and Bob then sacrifice k bits of their shared key to compare them.
If a nonmatching bit is found, the reason is either an eavesdropper or a noisy channel.
Since the probability for each eavesdropped bit to be wrong is 1/4, they will detect
Eve's presence with the probability 1-(3/4)^k.""")



def bernstein_vazirani(n=6, linear=True):
    r"""Bernstein-Vazirani algorithm demo.

    Simulates the Bernstein-Vazirani algorithm :cite:`BV`, which, given a black box oracle
    implementing a linear Boolean function :math:`f_a(x) := a \cdot x`, returns the bit
    vector a (and thus identifies the function) with just a single oracle call.
    If the oracle function is not linear, the algorithm will fail.
    """
    # Ville Bergholm 2011-2012

    print('\n\n=== Bernstein-Vazirani algorithm ===\n')

    print('Using {0} qubits.'.format(n))
    dim = qubits(n)
    H = gate.walsh(n) # n-qubit Walsh-Hadamard gate
    N = 2 ** n

    def oracle(f):
        """Returns a unitary oracle for the Boolean function f(x), given as a truth table."""
        return (-1) ** f.reshape((-1, 1))

    def linear_func(a):
        r"""Builds the linear Boolean function f(x) = a \cdot x as a truth table."""
        dim = qubits(len(a))
        N = prod(dim)
        U = empty(N, dtype = int)
        for k in range(N):
            x = np.unravel_index(k, dim)
            U[k] = mod(dot(a, x), 2)
        return U

    # black box oracle encoding the Boolean function f (given as the diagonal)
    if linear:
        # linear f
        a = asarray(npr.rand(n) > 0.5, dtype = int)
        f = linear_func(a)
        print('\nUsing the linear function f_a(x) := dot(a, x), defined by the binary vector a = {0}.'.format(a))
    else:
        # general f
        f = asarray(npr.rand(N) > 0.5, dtype = int)
        # special case: not(linear)
        #a = asarray(npr.rand(n) > 0.5, dtype = int)
        #f = 1-linear_func(a)
        print('\nNonlinear function f:\n{0}.'.format(f))
    U_oracle = oracle(f)

    # start with all-zero state
    s = state(0, dim)
    # initial superposition
    s = s.u_propagate(H)
    # oracle phase flip
    s.data = U_oracle * s.data
    # final Hadamards
    s = s.u_propagate(H)

    plt.figure()
    s.plot()
    title = 'Bernstein-Vazirani algorithm, '
    if linear:
        title += 'linear oracle'
    else:
        title += 'nonlinear oracle (fails)'
    plt.title(title)

    p, res = s.measure()
    # measured binary vector
    b = asarray(np.unravel_index(res, dim))
    print('\nMeasured binary vector b = {0}.'.format(b))
    if not(linear):
        g = linear_func(b)
        print('\nCorresponding linear function g_b:\n{0}.'.format(g))
        print('\nNormalized Hamming distance: |f - g_b| = {0}.'.format(sum(abs(f-g)) / N))

    return p



def grover_search(n=8):
    """Grover search algorithm demo.

    Simulate the Grover search algorithm :cite:`Grover` formulated using amplitude amplification
    in a system of n qubits.
    """
    # Ville Bergholm 2009-2010

    print('\n\n=== Grover search algorithm ===\n')

    A = gate.walsh(n) # Walsh-Hadamard gate for generating uniform superpositions
    N = 2 ** n # number of states

    sol = npr.randint(N)
    reps = int(pi/(4*asin(sqrt(1/N))))

    print('Using {0} qubits.'.format(n))
    print('Probability maximized by {0} iterations.'.format(reps))
    print('Correct solution: {0}'.format(sol))

    # black box oracle capable of recognizing the correct answer (given as the diagonal)
    # TODO an oracle that actually checks the solutions by computing (using ancillas?)
    U_oracle = ones((N, 1))
    U_oracle[sol, 0] = -1

    U_zeroflip = ones((N, 1))
    U_zeroflip[0, 0] = -1

    s = state(0, qubits(n))

    # initial superposition
    s = s.u_propagate(A)

    # Grover iteration
    for _ in range(reps):
        # oracle phase flip
        s.data = -U_oracle * s.data

        # inversion about the mean
        s = s.u_propagate(A.ctranspose())
        s.data = U_zeroflip * s.data
        s = s.u_propagate(A)

    p, res = s.measure()
    print('\nMeasured {0}.'.format(res))
    return p



def markov_decoherence(T1, T2, B=None):
    """Markovian decoherence demo.

    Given decoherence times T1 and T2, creates a markovian bath B
    and a coupling operator D which reproduce them on a single-qubit system.

    """
    # Ville Bergholm 2009-2011

    from . import markov
    print('\n\n=== Markovian decoherence in a qubit ===\n')

    omega0 = 2*pi* 1e9 # Hz
    T = 1 # K
    delta = 3 + 3*npr.rand() # qubit energy splitting (GHz)

    # setup the bath
    if B == None:
        B = markov.bath('ohmic', omega0, T) # defaults

    # find the correct qubit-bath coupling
    H, D = B.fit(delta, T1*omega0, T2*omega0)
    L = markov.superop(H, D, B)
    t = linspace(0, 10, 200)

    # T1 demo
    eq = 1 / (1 + exp(delta * B.scale)) # equilibrium rho_11
    s = state('1') # qubit in the |1> state
    out = s.propagate(L, t, lambda x, h: x.ev(p1))
    plt.figure()
    plt.subplot(2, 1, 1)
    plt.plot(t, out, 'r-', t, eq +(1-eq)*exp(-t/(T1*omega0)), 'b-.', [0, t[-1]], [eq, eq], 'k:', linewidth = 2)
    plt.xlabel('$t \\omega_0$')
    plt.ylabel('probability')
    plt.axis([0, t[-1], 0, 1])
    plt.title('$T_1$: relaxation')
    plt.legend(('simulated $P_1$', '$P_{1}^{eq} +(1-P_{1}^{eq}) \\exp(-t/T_1)$'))

    # T2 demo
    s = state('0')
    s = s.u_propagate(R_y(pi/2)) # rotate to (|0>+|1>)/sqrt(2)
    out = s.propagate(L, t, lambda x, h: x.u_propagate(R_y(-pi/2)).ev(p0))
    plt.subplot(2, 1, 2)
    plt.plot(t, out, 'r-', t, 0.5*(1+exp(-t/(T2*omega0))), 'b-.', linewidth = 2)
    plt.xlabel('$t \omega_0$')
    plt.ylabel('probability')
    plt.axis([0, t[-1], 0, 1])
    plt.title('$T_2$: dephasing')
    plt.legend(('simulated $P_0$', '$\\frac{1}{2} (1+\\exp(-t/T_2))$'))
    return H, D, B



def nmr_sequences(seqs=None, titles=None):
    """NMR control sequences demo.

    Compares the performance of different single-qubit NMR control
    sequences in the presence of systematic errors.
    Plots the fidelity of each control sequence as a function
    of both off-resonance error f and fractional pulse length error g.

    Reproduces fidelity plots in :cite:`Cummins`.
    """
    # Ville Bergholm 2006-2014

    from . import seq

    print('\n\n=== NMR control sequences for correcting systematic errors ===\n')

    if seqs == None:
        seqs = [seq.nmr([[pi, 0]]), seq.corpse(pi), seq.scrofulous(pi), seq.bb1(pi)]
        titles = [r'Plain $\pi$ pulse', 'CORPSE', 'SCROFULOUS', 'BB1']
    elif titles == None:
        titles = ['User-given seq'] * len(seqs)

    psi = state('0') # initial state
    f = arange(-1, 1, 0.05)
    g = arange(-1, 1, 0.08)
    nf = len(f)
    ng = len(g)

    def helper(ax, s_error, title):
        """Apply the sequence on the state psi, plot the evolution."""
        out, _ = seq.propagate(psi, s_error, out_func = state.bloch_vector)
        ax.set_title(title)
        plot.state_trajectory(out, ax = ax)


    for tt, ss in zip(titles, seqs):
        fig = plt.figure()
        gs = GridSpec(2, 2)
        U = seq.seq2prop(ss) # target propagator

        # The two systematic error types we are interested here can be
        # incorporated into the control sequence.
        #==================================================
        s_error = deepcopy(ss)
        s_error['A'] = ss['A'] + 0.1 * 0.5j * sz  # off-resonance error (constant \sigma_z drift term)
        ax = fig.add_subplot(gs[0, 0], projection = '3d')
        helper(ax, s_error, tt + ' evolution, off-resonance error')

        #==================================================
        s_error = deepcopy(ss)
        s_error['tau'] = ss['tau'] * 1.1  # proportional pulse length error
        ax = fig.add_subplot(gs[1, 0], projection = '3d')
        helper(ax, s_error, tt + ' evolution, pulse length error')

        #==================================================
        s_error = deepcopy(ss)
        fid = empty((ng, nf))

        def u_fidelity(a, b):
            """fidelity of two unitary rotations, [0,1]"""
            return 0.5 * abs(trace(dot(a.conj().transpose(), b)))

        for u in range(nf):
            s_error['A'] = ss['A'] + f[u] * 0.5j * sz  # off-resonance error
            for v in range(ng):
                s_error['tau'] = ss['tau'] * (1 + g[v])  # proportional pulse length error
                fid[v, u] = u_fidelity(U, seq.seq2prop(s_error))

        ax = fig.add_subplot(gs[:, 1])
        X, Y = meshgrid(f, g)
        ax.contour(X, Y, 1-fid)
        #ax.surf(X, Y, 1-fid)
        ax.set_xlabel('Off-resonance error')
        ax.set_ylabel('Pulse length error')
        ax.set_title(tt + ' fidelity')
        plt.show()



def phase_estimation(t, U, s, implicit=False):
    r"""Quantum phase estimation algorithm.

    Estimate an eigenvalue of the unitary :class:`lmap` U using t qubits,
    starting from the state s.

    Returns the state of the index register after the phase estimation
    circuit, but before final measurement.

    To get a result accurate to n bits with probability :math:`\ge 1 -\epsilon`,
    choose

    .. math::

       t \ge n + \left\lceil \log_2\left(2 +\frac{1}{2 \epsilon}\right) \right \rceil.

    See :cite:`Cleve`, :cite:`NC` chapter 5.2.
    """
    # Ville Bergholm 2009-2010

    T = 2 ** t
    S = U.data.shape[0]

    # index register in uniform superposition
    #reg_t = u_propagate(state(0, qubits(t)), gate.walsh(t)) # use Hadamards
    reg_t = state(ones(T) / sqrt(T), qubits(t)) # skip the Hadamards

    # state register (ignore the dimensions)
    reg_s = state(s, S)

    # full register
    reg = reg_t.tensor(reg_s)

    # controlled unitaries
    for k in range(t):
        ctrl = -ones(t)
        ctrl[k] = 1
        temp = gate.controlled(U ** (2 ** (t-1-k)), ctrl)
        reg = reg.u_propagate(temp)
    # from this point forward the state register is not used anymore

    if implicit:
        # save memory and CPU: make an implicit measurement of the state reg, discard the results
        _, _, reg = reg.measure(t, do = 'D')
        #print('Implicit measurement of state register: {0}\n', res)
    else:
        # more expensive computationally: trace over the state register
        reg = reg.ptrace((t,))

    # do an inverse quantum Fourier transform on the index reg
    QFT = gate.qft(qubits(t))
    return reg.u_propagate(QFT.ctranspose())



def phase_estimation_precision(t, U, u=None):
    """Quantum phase estimation demo.

    Estimate an eigenvalue of unitary operator U using t bits, starting from the state u.
    Plots and returns the probability distribution of the resulting t-bit approximations.
    If u is not given, use a random eigenvector of U.

    Uses :func:`phase_estimation`.
    """
    # Ville Bergholm 2009-2010

    print('\n\n=== Phase estimation ===\n\n')

    # find eigenstates of the operator
    N = U.shape[0]
    d, v = eig(U)
    if u == None:
        u = state(v[:, 0], N) # exact eigenstate

    print('Use {0} qubits to estimate the phases of the eigenvalues of a U({1}) operator.\n'.format(t, N))
    p = phase_estimation(t, lmap(U), u).prob()
    T = 2 ** t
    x = arange(T) / T
    w = 0.8 / T

    # plot probability distribution
    plt.figure()
    plt.hold(True)
    plt.bar(x, p, width = w) # TODO align = 'center' ???
    plt.xlabel(r'phase / $2\pi$')
    plt.ylabel('probability')
    plt.title('Phase estimation')
    #axis([-1/(T*2), 1-1/(T*2), 0, 1])

    # compare to correct answer
    target = angle(d) / (2*pi) + 1
    target -= floor(target)
    plt.plot(target, 0.5*max(p)*ones(len(target)), 'mo')

    plt.legend(('Target phases', 'Measurement probability distribution'))
    return p



def qft_circuit(dim=(2, 3, 3, 2)):
    r"""Quantum Fourier transform circuit demo.

    Simulate the quadratic QFT circuit construction.
    dim is the dimension vector of the subsystems.

    NOTE: If dim is not palindromic the resulting circuit also
    reverses the order of the dimensions in the SWAP cascade.

    .. math::

       U |x_1, x_2, \ldots, x_n\rangle
       = \frac{1}{\sqrt{d}} \sum_{k_i} |k_n,\ldots, k_2, k_1\rangle \exp\left(i 2 \pi \left(\sum_{r=1}^n k_r 0.x_r x_{r+1}\ldots x_n\right)\right)
       = \frac{1}{\sqrt{d}} \sum_{k_i} |k_n,\ldots, k_2, k_1\rangle \exp\left(i 2 \pi 0.x_1 x_2 \ldots x_n \left(\sum_{r=1}^n d_1 d_2 \cdots d_{r-1} k_r \right)\right).
    """
    # Ville Bergholm 2010-2011

    print('\n\n=== Quantum Fourier transform using a quadratic circuit ===\n')
    print('Subsystem dimensions: {0}'.format(dim))

    def Rgate(d):
        r"""R = \sum_{xy} exp(i*2*pi * x*y/prod(dim)) |xy><xy|"""
        temp = kron(arange(d[0]), arange(d[-1])) / prod(d)
        return gate.phase(2 * pi * temp, (d[0], d[-1]))

    n = len(dim)
    U = gate.id(dim)

    for k in range(n):
        H = gate.qft((dim[k],))
        U = gate.single(H, k, dim) * U
        for j in range(k+1, n):
            temp = Rgate(dim[k : j+1])
            U = gate.two(temp, (k, j), dim) * U

    for k in range(n // 2):
        temp = gate.swap(dim[k], dim[n-1-k])
        U = gate.two(temp, (k, n-1-k), dim) * U

    err = norm(U.data - gate.qft(dim).data)
    print('Error: {0}'.format(err))
    return U



def quantum_channels(p=0.3):
    """Visualization of simple one-qubit channels.

    Visualizes the effect of different quantum channels on a qubit using the Bloch sphere representation.
    """
    # Ville Bergholm 2009-2012

    print('\n\n=== Quantum channels ===\n')

    I = eye(2)
    E_bitflip      = [sqrt(1-p)*I, sqrt(p)*sx]
    E_phaseflip    = [sqrt(1-p)*I, sqrt(p)*sz]
    E_bitphaseflip = [sqrt(1-p)*I, sqrt(p)*sy]
    E_depolarize   = [sqrt(1-3*p/4)*I, sqrt(p)*sx/2, sqrt(p)*sy/2, sqrt(p)*sz/2]
    #t = arcsin(sqrt(gamma))
    t = pi/3
    E_amplitudedamp = [sqrt(p)*diag([1, cos(t)]), sqrt(p)*array([[0, sin(t)], [0, 0]]), sqrt(1-p)*diag([cos(t), 1]), sqrt(1-p)*array([[0, 0], [sin(t), 0]])]
    channels = [E_bitflip, E_phaseflip, E_bitphaseflip, E_depolarize, E_amplitudedamp]
    titles   = ['Bit flip', 'Phase flip', 'Bit and phase flip', 'Depolarization', 'Amplitude damping']

    X, Y, Z = plot.sphere()
    S = array([X, Y, Z])

    def present(ax, E, T):
        """Helper"""
        s = S.shape
        res = empty(s)

        for a in range(s[1]):
            for b in range(s[2]):
                temp = state.bloch_state(r_[1, S[:, a, b]])
                res[:, a, b] = temp.kraus_propagate(E).bloch_vector()[1:]  # skip the normalization

        plot.bloch_sphere(ax)
        ax.plot_surface(res[0], res[1], res[2], rstride = 1, cstride = 1, color = 'm', alpha = 0.2, linewidth = 0)
        ax.set_title(T)

    fig = plt.figure()
    plt.title('Quantum channels')
    n = len(channels)
    gs = GridSpec(2, int(ceil(n / 2)))
    for k in range(n):
        ax = fig.add_subplot(gs[k], projection = '3d')
        present(ax, channels[k], titles[k])

    plt.show()



def quantum_walk(steps=7, n=11, p=0.05, n_coin=2):
    """Quantum random walk demo.

    Simulates a 1D quantum walker controlled by a unitary quantum coin.
    On each step the coin is flipped and the walker moves either to the left
    or to the right depending on the result.

    After each step, the position of the walker is measured with probability p.
    p == 1 results in a fully classical random walk, whereas
    p == 0 corresponds to the "fully quantum" case.
    """
    # Ville Bergholm 2010-2011
    from .base import H

    print('\n\n=== Quantum walk ===\n')
    # initial state: coin shows heads, walker in center node
    coin   = state('0', n_coin)
    walker = state(n // 2, n)

    s = coin.tensor(walker).to_op(inplace = True)

    # translation operators (wrapping)
    left  = gate.mod_inc(-1, n)
    right = left.ctranspose()

    if n_coin == 2:
        # coin flip operator
        #C = R_x(pi / 2)
        C = H
        # shift operator: heads, move left; tails, move right
        S = kron(p0, left.data) +kron(p1, right.data)
    else:
        C = rand_U(n_coin)
        S = kron(diag([1, 0, 0]), left.data) +kron(diag([0, 1, 0]), right.data) +kron(diag([0, 0, 1]), eye(n))

    # propagator for a single step (flip + shift)
    U = dot(S, kron(C, eye(n)))

    # Kraus ops for position measurement
    M = []
    for k in range(n):
        temp = zeros((n, n))
        temp[k, k] = sqrt(p)
        M.append(kron(eye(n_coin), temp))

    # "no measurement"
    M.append(sqrt(1-p) * eye(n_coin * n))

    for k in range(steps):
        s = s.u_propagate(U)
        #s = s.kraus_propagate(M)
        s.data = p*diag(diag(s.data)) + (1-p)*s.data  # equivalent but faster...

    s = s.ptrace([0]) # ignore the coin
    plt.figure()
    s.plot()
    plt.title('Walker state after {0} steps'.format(steps))

    plt.figure()
    plt.bar(arange(n), s.prob(), width = 0.8)
    plt.title('Walker position probability distribution after {0} steps'.format(steps))
    return s


def qubit_and_resonator(d_r=30):
    """Qubit coupled to a microwave resonator demo.

    Simulates a qubit coupled to a microwave resonator.
    Reproduces plots from the experiment in :cite:`Hofheinz`.
    """
    # Ville Bergholm 2010-2014

    print('\n\n=== Qubit coupled to a single-mode microwave resonator ===\n')
    if d_r < 10:
        d_r = 10 # truncated resonator dim

    #omega0 = 1e9 # Energy scale/\hbar, Hz
    #T = 0.025 # K

    # omega_r = 2*pi* 6.570 # GHz, resonator angular frequency
    # qubit-resonator detuning Delta(t) = omega_q(t) -omega_r

    Omega     =  2*pi* 19e-3  # GHz, qubit-resonator coupling
    Delta_off = -2*pi* 463e-3 # GHz, detuning at off-resonance point
    Omega_q   =  2*pi* 0.5    # GHz, qubit microwave drive amplitude
    #  Omega << |Delta_off| < Omega_q << omega_r

    # decoherence times
    #T1_q = 650e-9 # s
    #T2_q = 150e-9 # s
    #T1_r = 3.5e-6 # s
    #T2_r = 2*T1_r # s

    # TODO heat bath and couplings
    #bq = markov.bath('ohmic', omega0, T)
    #[H, Dq] = fit(bq, Delta_off, T1_q, T2_q)
    #[H, Dr] = fit_ho(bq, ???, T1_r, T2_r???)
    #D = kron(eye(d_r), D) # +kron(Dr, qit.I)

    #=================================
    # Hamiltonians etc.

    # qubit raising and lowering ops
    sp = 0.5*(sx -1j*sy)
    sm = sp.conj().transpose()

    # resonator annihilation op
    a = boson_ladder(d_r)
    # resonator identity op
    I_r = eye(d_r)

    # qubit H
    Hq = kron(I_r, dot(sp, sm))
    # resonator H
    #Hr = kron(a'*a, I_q)

    # coupling H, rotating wave approximation
    Hint = Omega/2 * (kron(a, sp) +kron(a.conj().transpose(), sm))
    # microwave control H
    HOq = lambda ampl, phase: kron(I_r, ampl * 0.5 * Omega_q * (cos(phase) * sx + sin(phase) * sy))
    #Q = ho.position(d_r)
    #P = ho.momentum(d_r)
    #HOr = lambda ampl, phase: kron(ampl*Omega_r/sqrt(2)*(cos(phase)*Q +sin(phase)*P), qit.I)

    # system Hamiltonian, in rotating frame defined by H0 = omega_r * (Hq + Hr)
    # D = omega_q - omega_r is the detuning between the qubit and the resonator
    H = lambda D, ampl, phase:  D*Hq + Hint + HOq(ampl, phase)

    # readout: qubit in excited state?
    readout = lambda s, h: s.ev(kron(I_r, p1))

    s0 = state(0, (d_r, 2)) # resonator + qubit, ground state


    #=================================
    # Rabi test

    t = linspace(0, 500, 100)
    detunings = linspace(0, 2*pi* 40e-3, 100) # GHz
    out = empty((len(detunings), len(t)))

    #L = markov.superop(H(0, 1, 0), D, bq)
    L = H(0, 1, 0)  # zero detuning, sx pulse
    for k, d in enumerate(detunings):
        s = s0.propagate(L, (2/Omega_q)*pi/2) # Q (pi pulse for the qubit)
        #LL = markov.superop(H(d, 0, 0), D, bq)
        LL = H(d, 0, 0)  # detuned propagation
        out[k, :] = s.propagate(LL, t, out_func = readout)

    plt.figure()
    plot.pcolor(out, t, detunings / (2*pi*1e-3))
    #plt.colorbar(orientation = 'horizontal')
    plt.xlabel('Interaction time $\\tau$ (ns)')
    plt.ylabel('Detuning, $\\Delta/(2\\pi)$ (MHz)')
    plt.title('One photon Rabi-swap oscillations between qubit and resonator, $P_e$')

    #figure
    #f = fft(out, [], 2)
    #pcolor(abs(fftshift(f, 2)))
    #=================================
    # state preparation

    def demolish_state(targ):
        """Convert a desired (possibly truncated) resonator state ket into a program for constructing that state.
        State preparation in reverse, uses approximate idealized Hamiltonians.
        """
        # Ideal H without interaction
        A = lambda D, ampl, phase: D*Hq +HOq(ampl, phase)

        # resonator ket into a full normalized qubit+resonator state
        n = len(targ)
        targ = state(r_[targ, zeros(d_r-n)], (d_r,)).normalize().tensor(state(0, (2,)))
        t = deepcopy(targ)

        n = n-1 # highest excited level in resonator
        prog = zeros((n, 4))
        for k in reversed(range(n)):
            # |k+1,g> to |k,e> 
            dd = targ.data.reshape(d_r, 2)
            prog[k, 3] = (angle(dd[k, 1]) -angle(dd[k+1, 0]) -pi/2 +2*pi) / -Delta_off
            targ = targ.propagate(A(-Delta_off, 0, 0), prog[k, 3]) # Z

            dd = targ.data.reshape(d_r, 2)
            prog[k, 2] = (2/(sqrt(k+1) * Omega)) * arctan2(abs(dd[k+1, 0]), abs(dd[k, 1]))
            targ = targ.propagate(-Hint, prog[k, 2]) # S

            # |k,e> to |k,g>
            dd = targ.data.reshape(d_r, 2)
            phi = angle(dd[k, 1]) -angle(dd[k, 0]) +pi/2
            prog[k, 1] = phi
            prog[k, 0] = (2/Omega_q) * arctan2(abs(dd[k, 1]), abs(dd[k, 0]))
            targ = targ.propagate(A(0, -1, phi), prog[k, 0]) # Q
        return prog, t

    def prepare_state(prog):
        """Prepare a state according to the program."""
        s = s0 # start with ground state
        #s = tensor(ho.coherent_state(0.5, d_r), state(0, 2)) # coherent state
        for k in prog:
            # Q, S, Z
            s = s.propagate(H(0, 1, k[1]), k[0]) # Q
            s = s.propagate(H(0, 0, 0), k[2]) # S
            s = s.propagate(H(Delta_off, 0, 0), k[3]) # Z
        return s


    #=================================
    # readout plot (not corrected for limited visibility)

    prog, dummy = demolish_state([0, 1, 0, 1]) # |1> + |3>
    s1 = prepare_state(prog)

    prog, dummy = demolish_state([0, 1, 0, 1j]) # |1> + i|3>
    s2 = prepare_state(prog)

    t = linspace(0, 350, 200)
    out1 = s1.propagate(H(0, 0, 0), t, readout)
    out2 = s2.propagate(H(0, 0, 0), t, readout)

    plt.figure()
    plt.plot(t, out1, 'b-', t, out2, 'r-')
    plt.xlabel('Interaction time $\\tau$ (ns)')
    plt.ylabel('$P_e$')
    plt.title('Resonator readout through qubit.')
    plt.legend(('$|1\\rangle + |3\\rangle$', '$|1\\rangle + i|3\\rangle$'))


    #=================================
    # Wigner spectroscopy

    if False:
        # "voodoo cat"
        targ = zeros(d_r)
        for k in range(0, d_r, 3):
            targ[k] = (2 ** k) / sqrt(factorial(k))
    else:
        targ = [1, 0, 0, exp(1j * pi * 3/8), 0, 0, 1]

    # calculate the pulse sequence for constructing targ
    prog, t = demolish_state(targ)
    s = prepare_state(prog)

    print('Trying to prepare the state')
    print(t)
    print('Fidelity of prepared state with target state: {0:g}'.format(fidelity(s, t)))
    print('Time required for state preparation: {0:g} ns'.format(np.sum(prog[:, [0, 2, 3]])))
    print('\nComputing the Wigner function...')
    s = s.ptrace((1,))
    W, a, b = ho.wigner(s, res = (80, 80), lim = (-2.5, 2.5, -2.5, 2.5))
    plt.figure()
    plot.pcolor(W, a, b, (-1, 1))
    plt.title('Wigner function $W(\\alpha)$')
    plt.xlabel('Re($\\alpha$)')
    plt.ylabel('Im($\\alpha$)')



def shor_factorization(N=9, cheat=False):
    """Shor's factorization algorithm demo.

    Simulates Shor's factorization algorithm, tries to factorize the integer N.
    If cheat is False, simulates the full algorithm. Otherwise avoids the quantum part.

    NOTE: This is a very computationally intensive quantum algorithm
    to simulate classically, and probably will not run for any
    nontrivial value of N (unless you choose to cheat, in which case
    instead of simulating the quantum part (implemented in :func:`find_order`)
    we use a more efficient classical algorithm for the order-finding).

    See :cite:`Shor`, :cite:`NC` chapter 5.3.
    """
    # Ville Bergholm 2010-2011

    def find_order_cheat(a, N):
        """Classical order-finding algorithm."""
        for r in range(1, N+1):
            if mod_pow(a, r, N) == 1:
                return r

    def mod_pow(a, x, N):
        """Computes mod(a^x, N) using repeated squaring mod N.
        x must be a positive integer.
        """
        X = numstr_to_array(bin(x)[2:]) # exponent in big-endian binary
        res = 1
        for b in reversed(X): # little-endian
            if b:
                res = mod(res*a, N)
            a = mod(a*a, N) # square a
        return res


    print('\n\n=== Shor\'s factorization algorithm ===\n')
    if cheat:
        print('(cheating)\n')

    # number of bits needed to represent mod N arithmetic:
    m = int(ceil(log2(N)))
    print('Trying to factor N = {0} ({1} bits).'.format(N, m))

    # maximum allowed failure probability for the quantum order-finding part
    epsilon = 0.25
    # number of index qubits required for the phase estimation
    t = 2*m +1 +int(ceil(log2(2 + 1 / (2 * epsilon))))
    print('The quantum order-finding subroutine will need {0} + {1} qubits.\n'.format(t, m))

    # classical reduction of factoring to order-finding
    while True:
        a = npr.randint(2, N) # random integer, 2 <= a < N
        print('Random integer: a = {0}'.format(a))
  
        p = gcd(a, N)
        if p != 1:
            # a and N have a nontrivial common factor p.
            # This becomes extremely unlikely as N grows.
            print('Lucky guess, we found a common factor!')
            break

        print('Trying to find the period of f(x) = a^x mod N')

        if cheat:
            # classical cheating shortcut
            r = find_order_cheat(a, N)
        else:
            while True:
                print('.')
                # ==== quantum part of the algorithm ====
                [s1, r1] = find_order(a, N, epsilon)
                [s2, r2] = find_order(a, N, epsilon)
                # =============  ends here  =============

                if gcd(s1, s2) == 1:
                    # no common factors
                    r = lcm(r1, r2)
                    break

        print('\n  =>  r = {0}'.format(r))

        # if r is odd, try again
        if gcd(r, 2) == 2:
            # r is even

            x = mod_pow(a, r // 2, N)
            if mod(x, N) != N-1:
                # factor found
                p = gcd(x-1, N) # ok?
                if p == 1:
                    p = gcd(x+1, N) # no, try this
                break
            else:
                print('a^(r/2) = -1 (mod N), try again...\n')
        else:
            print('r is odd, try again...\n')

    print('\nFactor found: {0}'.format(p))
    return p


def find_order(a, N, epsilon=0.25):
    """Quantum order-finding subroutine.

    Finds the period of the function f(x) = a^x mod N.
    epsilon is the maximum allowed failure probability for the subroutine.

    Returns r, s where r/s approximates period/T.

    Uses :func:`phase_estimation`.

    See :cite:`Shor`, :cite:`NC` chapter 5.3.
    """
    # number of bits needed to represent mod N arithmetic:
    m = int(ceil(log2(N)))
    # number of index qubits required for the phase estimation
    t = 2*m +1 +int(ceil(log2(2 + 1 / (2 * epsilon))))

    T = 2 ** t # index register dimension
    M = 2 ** m # state register dimension

    # applying f(x) is equivalent to the sequence of controlled modular multiplications in phase estimation
    U = gate.mod_mul(a, M, N)

    # state register initialized in the state |1>
    st = state(1, M)

    # run the phase estimation algorithm
    reg = phase_estimation(t, U, st, implicit = True) # use implicit measurement to save memory

    # measure index register
    dummy, num = reg.measure()

    def find_denominator(x, y, max_den):
        r"""
        Finds the denominator s for r/s \approx x/y such that s < max_den
        using the convergents of the continued fraction representation

        .. math::

           \frac{x}{y} = a_0 +\frac{1}{a_1 +\frac{1}{a_2 +\ldots}}

        We use floor and mod here, which could be both efficiently implemented using
        the classical Euclidean algorithm.
        """
        d_2 = 1
        d_1 = 0
        while True:
            a = x // y # integer part == a_n
            temp = a*d_1 +d_2 # n:th convergent denumerator d_n = a_n*d_{n-1} +d_{n-2}
            if temp >= max_den:
                break

            d_2 = d_1
            d_1 = temp
            temp = mod(x, y)  # x - a*y # subtract integer part
            if temp == 0:
            #if (temp/y < 1 / (2*max_den ** 2))
                break  # continued fraction terminates here, result is exact

            # invert the remainder (swap numerator and denominator)
            x = y
            y = temp
        return d_1

    # another classical part
    s = find_denominator(num, T, T+1)
    r = (num * s) // T
    return r, s


def superdense_coding(d=2):
    """Superdense coding demo.

    Simulate Alice sending two d-its of information to Bob using 
    a shared EPR qudit pair.

    """
    # Ville Bergholm 2010-2011

    print('\n\n=== Superdense coding ===\n')

    H   = gate.qft(d)        # qft (generalized Hadamard) gate
    add = gate.mod_add(d, d) # modular adder (generalized CNOT) gate
    I   = gate.id(d)

    dim = (d, d)

    # EPR preparation circuit
    U = add * tensor(H, I)

    print('Alice and Bob start with a shared EPR pair:')
    reg = state('00', dim).u_propagate(U)
    print(reg)

    # two random d-its
    a = floor(d * npr.rand(2)).astype(int)
    print('\nAlice wishes to send two d-its of information (d = {0}) to Bob: a = {1}.'.format(d, a))

    Z = H * gate.mod_inc(a[0], d) * H.ctranspose()
    X = gate.mod_inc(-a[1], d)

    print('Alice encodes the d-its to her half of the EPR pair using local transformations,')
    reg = reg.u_propagate(tensor(Z*X, I))
    print(reg)

    print('\nand sends it to Bob. He then disentangles the pair,')
    reg = reg.u_propagate(U.ctranspose())
    print(reg)

    _, b = reg.measure()
    b = array(np.unravel_index(b, dim))
    print('\nand measures both qudits in the computational basis, obtaining the result  b = {0}.'.format(b))

    if all(a == b):
        print('The d-its were transmitted succesfully.')
    else:
        raise RuntimeError('Should not happen.')


def teleportation(d=2):
    """Quantum teleportation demo.

    Simulate the teleportation of a d-dimensional qudit from Alice to Bob.

    """
    # Ville Bergholm 2009-2011

    print('\n\n=== Quantum teleportation ===\n')

    H   = gate.qft(d)        # qft (generalized Hadamard) gate
    add = gate.mod_add(d, d) # modular adder (generalized CNOT) gate
    I   = gate.id(d)

    dim = (d, d)

    # EPR preparation circuit
    U = add * tensor(H, I)

    print('Alice and Bob start with a shared EPR pair:')
    epr = state('00', dim).u_propagate(U)
    print(epr)

    print('\nAlice wants to transmit this payload to Bob:')
    payload = state('0', d).u_propagate(rand_SU(d)).fix_phase()
    # choose a nice global phase
    print(payload)

    print('\nThe total |payload> \otimes |epr> register looks like')
    reg = payload.tensor(epr)
    print(reg)

    print('\nNow Alice entangles the payload with her half of the EPR pair,')
    reg = reg.u_propagate(tensor(U.ctranspose(), I))
    print(reg)

    _, b, reg = reg.measure((0, 1), do = 'C')
    b = array(np.unravel_index(b, dim))
    print('\nand measures her qudits, getting the result {0}.'.format(b))
    print('She then transmits the two d-its to Bob using a classical channel.')
    print('Since Alice\'s measurement has unentangled the state,')
    print('Bob can ignore her qudits. His qudit now looks like')
    reg_B = reg.ptrace((0, 1)).to_ket() # pure state
    print(reg_B)

    print('\nUsing the two classical d-its of data Alice sent him,')
    print('Bob performs a local transformation on his half of the EPR pair.')
    Z = H * gate.mod_inc(b[0], d) * H.ctranspose()
    X = gate.mod_inc(-b[1], d)
    reg_B = reg_B.u_propagate(Z*X).fix_phase()
    print(reg_B)

    ov = fidelity(payload, reg_B)
    print('\nThe overlap between the resulting state and the original payload state is |<payload|B>| = {0}'.format(ov))
    if abs(ov-1) > tol:
        raise RuntimeError('Should not happen.')
    else:
        print('The payload state was succesfully teleported from Alice to Bob.')
    return reg_B, payload



def tour():
    """Guided tour to the quantum information toolkit.

    """
    # Ville Bergholm 2009-2014

    print('This is the guided tour for the Quantum Information Toolkit.')
    print("It should be run in the interactive mode, 'ipython --pylab'.")
    print('Between examples, press any key to proceed to the next one.')

    def pause():
        plt.waitforbuttonpress()

    temp = plt.figure()
    teleportation()
    pause()
    superdense_coding()
    pause()
    plt.close(temp)

    adiabatic_qc_3sat(5, 25)
    pause()

    phase_estimation_precision(5, rand_U(4))
    plt.title('Phase estimation, eigenstate')
    phase_estimation_precision(5, rand_U(4), state(0, [4]))
    plt.title('Phase estimation, random state')
    pause()

    nmr_sequences()
    pause()

    quantum_channels()
    pause()

    bernstein_vazirani(6, linear = True)
    bernstein_vazirani(6, linear = False)
    pause()

    grover_search(6)
    pause()

    # shor_factorization(9) # TODO need to use sparse matrices to get even this far(!)
    shor_factorization(311*269, cheat = True)
    pause()

    bb84(40)
    pause()

    markov_decoherence(7e-10, 1e-9)
    pause()

    qubit_and_resonator(20)
    pause()

    qft_circuit((2, 3, 2))

    #quantum_walk()



# -*- coding: utf-8 -*-
"Unit tests for qit.lmap"
# Ville Bergholm 2009-2014

import unittest
from numpy.random import rand, randn
from numpy.linalg import norm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base  import tol
from qit.lmap  import lmap, tensor
from qit.utils import mkron


def randn_complex(*arg):
    "Returns an array of random complex numbers, normally distributed."
    return randn(*arg) +1j*randn(*arg)


class LmapConstructorTest(unittest.TestCase):
    def test_constructor(self):
        "Test lmap.__init__"

        # 1D
        s = lmap(randn(4, 1))             # ket vector
        s = lmap(randn(4))                # ket vector as 1D array
        s = lmap(randn(1, 3))             # bra vector
        s = lmap(randn(3), ((1,), None))  # bra vector as 1D array
        # 2D
        s = lmap(randn(4, 5), ((2, 2), None))    # input dims inferred
        s = lmap(randn(3, 6), (None, (3, 2)))    # output dims inferred
        s = lmap(randn(6, 6), ((2, 3), (3, 2)))  # all dims given
        temp = lmap(randn_complex(4, 4))         # both dims inferred

        # copy constructor
        s = lmap(temp, ((1, 4), (2, 2))) # dims reinterpreted
        s = lmap(temp, ((1, 4), None))   # input dims kept

        # bad inputs
        self.assertRaises(ValueError, lmap, rand(2,3), ((2,), (2, 3)))  # dimension mismatch
        self.assertRaises(ValueError, lmap, rand(2, 2, 2)) # bad array dimesion (3)



class LmapMethodTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_methods(self):
        ### reorder, tensor
        idim = (2, 5, 3)
        odim = (4, 3, 2)
        # generate random local maps
        parts = []
        for k in range(len(idim)):
            parts.append(lmap(rand(odim[k], idim[k])))
        T1 = tensor(*tuple(parts))
        # permute them
        perms = [(2, 0, 1), (1, 0, 2), (2, 1, 0)]
        for p in perms:
            T2 = T1.reorder((p, p))
            tup = (parts[i] for i in p)
            self.assertAlmostEqual((tensor(*tup) -T2).norm(), 0, delta=tol)

        ignore = """
        print(repr(b * c))
        print(repr(tensor(a, b)))
        """



if __name__ == '__main__':
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
"""
Utilities (:mod:`qit.utils`)
============================

.. currentmodule:: qit.utils

This module contains utility functions which do not logically fit anywhere else.


Mathematical utilities
----------------------

.. autosummary::

   gcd
   lcm
   majorize


Matrix functions
----------------

.. autosummary::

   comm
   acomm
   mkron
   rank
   orth
   nullspace
   nullspace_hermitian
   projector
   eighsort
   expv


Random matrices
---------------

.. autosummary::

   rand_hermitian
   rand_U
   rand_SU
   rand_U1
   rand_pu
   rand_positive
   rand_SL


Superoperators
--------------

.. autosummary::

   vec
   inv_vec
   lmul
   rmul
   lrmul
   superop_lindblad   
   superop_fp

Physics
-------

.. autosummary::

   angular_momentum
   boson_ladder
   fermion_ladder


Bases, decompositions
---------------------

.. autosummary::

   spectral_decomposition
   gellmann
   tensorbasis


Miscellaneous
-------------

.. autosummary::

   assert_o
   copy_memoize
   op_list
   qubits
   R_nmr
   R_x
   R_y
   R_z
"""
# Ville Bergholm 2008-2012

from __future__ import division, absolute_import, print_function, unicode_literals

from copy import deepcopy

import numpy as np
from numpy import (array, mat, dtype, empty, zeros, ones, eye, prod, sqrt, exp,
    dot, sort, diag, trace, kron, pi, r_, c_, inf, isscalar, floor, ceil, log, log10, vdot)
from numpy.random import rand, randn
from numpy.linalg import qr, det, eigh, eigvalsh
from scipy.linalg import expm, norm, svd, svdvals

from .base import sx, sy, sz, tol

__all__ = ['assert_o', 'copy_memoize',
           'gcd', 'lcm', 'majorize',
           'comm', 'acomm', 'mkron', 'projector', 'eighsort', 'expv',
           'rank', 'orth', 'nullspace', 'nullspace_hermitian',
           'rand_hermitian', 'rand_U', 'rand_SU', 'rand_U1', 'rand_pu', 'rand_positive', 'rand_SL', 
           'vec', 'inv_vec', 'lmul', 'rmul', 'lrmul', 'superop_lindblad', 'superop_fp',
           'angular_momentum', 'boson_ladder', 'fermion_ladder',
           'R_nmr', 'R_x', 'R_y', 'R_z',
           'spectral_decomposition',
           'gellmann', 'tensorbasis',
           'op_list',
           'qubits']


# the functions in this module return ndarrays, not lmaps, for now


# internal utilities

def _warn(s):
    """Prints a warning."""
    print('Warning: ' + s)


def assert_o(actual, desired, tolerance):
    """Octave-style assert."""
    if abs(actual - desired) > tolerance:
        raise AssertionError


def copy_memoize(func):
    """Memoization decorator for functions with immutable args, returns deep copies."""
    cache = {}
    def wrapper(*args):
        """Nonsense, this is an election year."""
        if args in cache:
            value = cache[args]
        else:
            value = func(*args)
            cache[args] = value

        return deepcopy(value)

    # so that the help system still works
    wrapper.__name__ = func.__name__
    wrapper.__doc__  = func.__doc__
    return wrapper



# math functions

def gcd(a, b):
    """Greatest common divisor.

    Uses the Euclidean algorithm.
    """
    while b:
        a, b = b, a % b
    return a


def lcm(a, b):
    """Least common multiple."""
    return a * (b // gcd(a, b))


def rank(A, tol=None):
    """Matrix rank."""
    s = svdvals(A)
    if tol == None:
        tol = max(A.shape) * np.amax(s) * np.finfo(A.dtype).eps
    return np.sum(s > tol, dtype=int)


def orth(A, tol=None, kernel=False, spectrum=False):
    """Construct an orthonormal basis for the range of A using SVD

    Parameters
    ----------
    A : array, shape (M, N)

    Returns
    -------
    Q : array, shape (M, K)
        Orthonormal basis for the range of A.
        K = effective rank of A, as determined by tolerance

    Adaptation of scipy.linalg.orth with optional absolute tolerance.
    """
    U, s, Vh = svd(A, full_matrices=False)
    if tol == None:
        tol = max(A.shape) * np.amax(s) * np.finfo(A.dtype).eps
    num = np.sum(s > tol, dtype=int)

    if not kernel:
        # range of A
        ret = U[:, :num]
    else:
        # kernel of A
        ret = Vh[num:, :].conj().transpose()

    if spectrum:
        return ret, s
    else:
        return ret


def nullspace(A, tol=None, spectrum=False):
    """Kernel of a matrix.

    Given a matrix A (and optionally a tolerance tol), returns a basis
    for the kernel (null space) of A in the columns of Z.
    """
    # the computation is almost the same as in orth, so...
    return orth(A, tol, kernel=True, spectrum=spectrum)


def nullspace_hermitian(A, tol=None):
    r"""Kernel of a superoperator matrix restricted to the Hermitian subspace.

    Solves the intersection of the kernel of the superop A and the Hermitian subspace.
    A maps d*d matrices to whatever.

    Singular values <= tol are considered zero.
    """
    # Hermitian and antihermitian (orthogonal) real subspaces: V = H \oplus A, h \in H, a \in A
    # If G = A'*A is either block-diagonal or block-antidiagonal wrt these, 
    # x = h+a, Gx = \lambda x implies that Gh = \lambda h and Aa = \lambda a.
    # Hence...
    # Ax := [Q, x] is block-antidiagonal if Q is hermitian
    
    # Ville Bergholm 2011-2012

    def to_real(C):
        """Represents a complex linear map in a real vector space.
        Corresponding transformation for column vectors: x_R = r_[x.real, x.imag]
        """
        return r_[c_[C.real, -C.imag], c_[C.imag, C.real]]

    D = A.shape[1]
    d = int(sqrt(D))  # since A is a superop

    # Hermitian basis.
    # reshape uses row-major order, so technically we would need transpose each matrix,
    # but since they're Hermitian and orthonormal either way it doesn't really matter.
    V = gellmann(d).reshape((D-1, D))
    # Add identity
    V = c_[eye(d).flatten() / sqrt(d), V.transpose()]
    U = r_[V.real, V.imag] # same, except now in a real vector space

    # find A restricted to the Hermitian (real) subspace H
    AH = dot(to_real(A), U)

    # solve the kernel
    Z, spectrum = nullspace(AH, tol, spectrum=True) # null space (kernel) of AH

    # columns of C: complex orthogonal vectors spanning the kernel (with real coefficients)
    C = dot(V, Z)
    return C, spectrum


def projector(v):
    """Projector corresponding to vector v."""
    return np.outer(v, v.conj())


def eighsort(A):
    """Returns eigenvalues and eigenvectors of a Hermitian matrix, sorted in nonincreasing order."""
    d, v = eigh(A)
    ind = d.argsort()[::-1]  # nonincreasing real part
    return d[ind], v[:, ind]


def comm(A, B):
    """Array commutator.
    
    Returns [A, B] := A*B - B*A 
    """
    return dot(A, B) - dot(B, A)


def acomm(A, B):
    """Array anticommutator.
    
    Returns {A, B} := A*B + B*A 
    """
    return dot(A, B) + dot(B, A)


def expv(t, A, v, tol=1.0e-7, m=30, iteration='arnoldi'):
    r"""Multiply a vector by an exponentiated matrix.

    Approximates :math:`exp(t A) v` using a Krylov subspace technique.
    Efficient for large sparse matrices.
    The basis for the Krylov subspace is constructed using either Arnoldi or Lanczos iteration.

    Input:
    t           vector of nondecreasing time instances >= 0
    A           n*n matrix (usually sparse) (as an (n,n)-shaped ndarray)
    v           n-dimensional vector (as an (n,)-shaped ndarray)
    tol         tolerance
    m           Krylov subspace dimension, <= n
    iteration   'arnoldi' or 'lanczos'. Lanczos is faster but requires a Hermitian A.

    Output:
    W       result matrix, :math:`W[i,:] \approx \exp(t[i] A) v`
    error   total truncation error estimate
    hump    :math:`\max_{s \in [0, t]}  \| \exp(s A) \|`

    Uses the sparse algorithm from :cite:`EXPOKIT`.
    """
    # Ville Bergholm 2009-2012

    # just in case somebody tries to use numpy.matrix instances here
    if isinstance(A, np.matrix) or isinstance(v, np.matrix):
        raise ValueError("A and v must be plain numpy.ndarray instances, not numpy.matrix.")

    n = A.shape[0]
    m = min(n, m)  # Krylov space dimension, m <= n 

    if isscalar(t):
        tt = array([t])
    else:
        tt = t

    a_norm = norm(A, inf)
    v_norm = norm(v)

    happy_tol  = 1.0e-7  # "happy breakdown" tolerance
    min_error = a_norm * np.finfo(float).eps # due to roundoff

    # step size control
    max_stepsize_changes = 10
    # safety factors
    gamma = 0.9
    delta = 1.2
    # initial stepsize
    fact = sqrt(2 * pi * (m + 1)) * ((m + 1) / exp(1)) ** (m + 1)

    def ceil_at_nsd(x, n = 2):
        temp = 10 ** (floor(log10(x))-n+1)
        return ceil(x / temp) * temp

    def update_stepsize(step, err_loc, r):
        step *= gamma  * (tol * step / err_loc) ** (1 / r)
        return ceil_at_nsd(step, 2)

    dt = dtype(complex)
    # TODO don't use complex matrices unless we have to: dt = result_type(t, A, v)

    # TODO shortcuts for Hessenberg matrix exponentiation?
    H = zeros((m+2, m+2), dt) # upper Hessenberg matrix for the Arnoldi process + two extra rows/columns for the error estimate trick
    H[m + 1, m] = 1           # never overwritten!
    V = zeros((n, m+1), dt)   # orthonormal basis for the Krylov subspace + one extra vector

    W = empty((len(tt), len(v)), dt)  # results
    t = 0  # current time
    beta = v_norm
    error = 0  # error estimate
    hump = [[v_norm, t]]
    #v_norm_max = v_norm  # for estimating the hump

    def iterate_lanczos(v, beta):
        """Lanczos iteration, for Hermitian matrices.
        Produces a tridiagonal H, cheaper than Arnoldi.

        Returns the number of basis vectors generated, and a boolean indicating a happy breakdown.
        NOTE that the we _must_not_ change global variables other than V and H here
        """
        # beta_0 and alpha_m are not used in H, beta_m only in a single position for error control 
        prev = 0
        for k in range(0, m):
            vk = (1 / beta) * v
            V[:, k] = vk  # store the now orthonormal basis vector
            # construct the next Krylov vector beta_{k+1} v_{k+1}
            v = dot(A, vk)
            H[k, k] = alpha = vdot(vk, v)
            v += -alpha * vk -beta * prev
            # beta_{k+1}
            beta = norm(v)
            if beta < happy_tol: # "happy breakdown": iteration terminates, Krylov approximation is exact
                return k+1, True
            if k == m-1:
                # v_m and one beta_m for error control (alpha_m not used)
                H[m, m-1] = beta
                V[:, m] = (1 / beta) * v
            else:
                H[k+1, k] = H[k, k+1] = beta
                prev = vk
        return m+1, False

    def iterate_arnoldi(v, beta):
        """Arnoldi iteration, for generic matrices.
        Produces a Hessenberg-form H.
        """
        V[:, 0] = (1 / beta) * v  # the first basis vector v_0 is just v, normalized
        for j in range(1, m+1):
            p = dot(A, V[:, j-1])  # construct the Krylov vector v_j
            # orthogonalize it with the previous ones
            for i in range(j):
                H[i, j-1] = vdot(V[:, i], p)
                p -= H[i, j-1] * V[:, i]
            temp = norm(p) 
            if temp < happy_tol: # "happy breakdown": iteration terminates, Krylov approximation is exact
                return j, True
            # store the now orthonormal basis vector
            H[j, j-1] = temp
            V[:, j] = (1 / temp) * p
        return m+1, False  # one extra vector for error control

    # choose iteration type
    iteration = iteration.lower()
    if iteration == 'lanczos':
        iteration = iterate_lanczos  # only works for Hermitian matrices!
    elif iteration == 'arnoldi':
        iteration = iterate_arnoldi
    else:
        raise ValueError("Only 'arnoldi' and 'lanczos' iterations are supported.")

    # loop over the time instances (which must be in increasing order)
    for kk in range(len(tt)):
        t_end = tt[kk]
        # initial stepsize
        # TODO we should inherit the stepsize from the previous interval
        r = m
        t_step = (1 / a_norm) * ((fact * tol) / (4 * beta * a_norm)) ** (1 / r)
        t_step = ceil_at_nsd(t_step, 2)

        while t < t_end:
            t_step = min(t_end - t, t_step)  # step at most the remaining distance

            # Arnoldi/Lanczos iteration, (re)builds H and V
            j, happy = iteration(v, beta)
            # now V^\dagger A V = H  (just the first m vectors, or j if we had a happy breakdown!)
            # assert(norm(dot(dot(V[:, :m].conj().transpose(), A), V[:, :m]) -H[:m,:m]) < tol)

            # error control
            if happy:
                # "happy breakdown", using j Krylov basis vectors
                t_step = t_end - t  # step all the rest of the way
                F = expm(t_step * H[:j, :j])
                err_loc = happy_tol
                r = m
            else:
                # no happy breakdown, we need the error estimate (using all m+1 vectors)
                av_norm = norm(dot(A, V[:, m]))
                # find a reasonable step size
                for k in range(max_stepsize_changes + 1):
                    F = expm(t_step * H)
                    err1 = beta * abs(F[m, 0])
                    err2 = beta * abs(F[m+1, 0]) * av_norm
                    if err1 > 10 * err2:  # quick convergence
                        err_loc = err2
                        r = m 
                    elif err1 > err2:  # slow convergence
                        err_loc = (err2 * err1) / (err1 - err2)
                        r = m 
                    else:  # asymptotic convergence
                        err_loc = err1
                        r = m-1
                    # should we accept the step?
                    if err_loc <= delta * tol * t_step:
                        break
                    if k >= max_stepsize_changes:
                        raise RuntimeError('Requested tolerance cannot be achieved in {0} stepsize changes.'.format(max_stepsize_changes))
                    t_step = update_stepsize(t_step, err_loc, r)

            # step accepted, update v, beta, error, hump
            v = dot(V[:, :j], beta * F[:j, 0])
            beta = norm(v)
            error += max(err_loc, min_error)
            #v_norm_max = max(v_norm_max, beta)

            t += t_step
            t_step = update_stepsize(t_step, err_loc, r)
            hump.append([beta, t])

        W[kk, :] = v

    hump = array(hump) / v_norm
    return W, error, hump



# random matrices

def rand_hermitian(n):
    """Random Hermitian n*n matrix.

    Returns a random Hermitian matrix of size n*n.
    NOTE: The randomness is not defined in any deeply meaningful sense.
    """
    # Ville Bergholm 2008-2009

    H = (rand(n, n) - 0.5) +1j*(rand(n, n) - 0.5)
    return H + H.conj().transpose() # make it Hermitian


def rand_U(n):
    """Random U(n) matrix.

    Returns a random unitary n*n matrix.
    The matrix is random with respect to the Haar measure.

    Uses the algorithm in :cite:`Mezzadri`.
    """
    # Ville Bergholm 2005-2014

    # sample the Ginibre ensemble, p(Z(i,j)) == 1/pi * exp(-abs(Z(i,j))^2),
    # p(Z) == 1/pi^(n^2) * exp(-trace(Z'*Z))
    Z = (randn(n, n) + 1j*randn(n, n)) / sqrt(2)

    # QR factorization
    Q, R = qr(Z)

    # eliminate multivaluedness in Q
    P = diag(R).copy()  # TODO remove copy() once we have numpy 1.10
    P /= abs(P)
    return dot(Q, diag(P))


def rand_SU(n):
    """Random SU(n) matrix.

    Returns a random special unitary n*n matrix.
    The matrix is random with respect to the Haar measure.
    """
    # Ville Bergholm 2005-2009

    U = rand_U(n)
    d = det(U) ** (1/n) # *exp(i*2*pi*k/n), not unique FIXME
    return U/d


def rand_U1(n):
    """Random diagonal unitary matrix.

    Returns a random diagonal unitary n*n matrix.
    The matrix is random with respect to the Haar measure.
    """
    # Ville Bergholm 2005-2009

    return diag(exp(2j * pi * rand(n)))


def rand_pu(n):
    r"""Random n-partition of unity.

    Returns the n-partition p, which is random with respect to
    the order-n Dirichlet distribution :math:`Dir(\alpha)`
    with :math:`\alpha = (1, 1, ..., 1)`.
    """
    # Ville Bergholm 2008-2012

    # Sample the Dirichlet distribution using n gamma-distributed
    # random vars. The (shared) scale parameter of the gamma pdfs is irrelevant,
    # and the shape parameters correspond to the Dirichlet \alpha params.
    # Furthermore, Gamma(x; 1,1) = exp(-x), so
    p = -log(rand(n))  # Gamma(1,1) -distributed
    return p / sum(p)  # Dir(1, 1, ..., 1) -distributed

    # TODO this would be a simpler choice, but what's the exact distribution?
    #p = sort(rand(n-1))  # n-1 points in [0,1]
    #d = sort(r_[p, 1] - r_[0, p])  # n deltas between points = partition of unity


def rand_positive(n):
    """Random n*n positive semidefinite matrix.

    Normalized to Tr(A) = 1.
    Since A has all-real eigenvalues, it is Hermitian by construction.
    """
    # Ville Bergholm 2008-2012

    d = rand_pu(n) # random partition of unity
    U = mat(rand_U(n)) # random unitary
    A = U.H * diag(d) * U
    return array((A + A.H) / 2)   # eliminate rounding errors
    # An alternative would be to use an inverse purification, but this would be expensive.


def rand_SL(n):
    """Random SL(n, C) matrix.

    Returns a random special linear n*n matrix.
    NOTE: The randomness is not defined in any deeply meaningful sense.
    """
    # Ville Bergholm 2011

    S = randn(n, n) +1j*randn(n, n)
    d = det(S) ** (1/n)
    return S/d



# superoperators

def vec(rho):
    """Flattens a matrix into a vector.

    Matrix rho is flattened columnwise into a column vector v.

    Used e.g. to convert state operators to superoperator representation.
    """
    # JDW 2009
    # Ville Bergholm 2009

    return rho.flatten('F')  # copy


def inv_vec(v, dim=None):
    r"""Reshapes a vector into a matrix.

    Given dim == (n, m), reshapes vector v (length n*m) into a matrix :math:`\rho` (shape == dim),
    using column-major ordering. If dim is not given, :math:`\rho` is assumed to be square.

    Used e.g. to convert state operators from superoperator representation
    to standard matrix representation.
    """
    # JDW 2009
    # Ville Bergholm 2009

    d = v.size
    if dim == None:
        # assume a square matrix
        n = sqrt(d)
        if floor(n) != n:
            raise ValueError('Length of vector v is not a squared integer.')
        dim = (n, n)
    else:
        if prod(dim) != d:
            raise ValueError('Dimensions n, m are not compatible with given vector v.')
    return v.reshape(dim, order='F').copy()


def lmul(L, q=None):
    r"""Superoperator equivalent for multiplying from the left.

    :param array L: matrix, shape (m, p)
    :param int q: The shape of :math:`\rho` is (p, q). If q is not given, :math:`\rho` is assumed square.

    Returns the superoperator that implements left multiplication
    of a vectorized matrix :math:`\rho` by the matrix L.

    .. math:: L \rho = \text{inv\_vec}(\text{lmul}(L) \text{vec}(\rho))
    """
    # Ville Bergholm 2009

    if q == None:
        q = L.shape[1]  # assume target is a square matrix
    return kron(eye(q), L)


def rmul(R, p=None):
    r"""Superoperator equivalent for multiplying from the right.

    :param array R: matrix, shape (q, r)
    :param int p: The shape of :math:`\rho` is (p, q). If p is not given, :math:`\rho` is assumed square.

    Returns the superoperator that implements right multiplication
    of a vectorized matrix :math:`\rho` by the matrix R.

    .. math:: \rho R = \text{inv\_vec}(\text{rmul}(R) \text{vec}(\rho))
    """
    # Ville Bergholm 2009

    if p == None:
        p = R.shape[0]  # assume target is a square matrix
    return kron(R.transpose(), eye(p))


def lrmul(L, R):
    r"""Superoperator equivalent for multiplying both from left and right.

    :param array L: matrix, shape (m, p)
    :param array R: matrix, shape (q, r)

    Returns the superoperator that implements left multiplication
    by the matrix L and right multiplication by the matrix R
    of a vectorized matrix :math:`\rho`.

    .. math:: L \rho R = \text{inv\_vec}(\text{lrmul}(L, R) \text{vec}(\rho))
    """
    # Ville Bergholm 2009-2011

    # L and R fix the shape of rho completely
    return kron(R.transpose(), L)


def superop_lindblad(A, H=None):
    r"""Liouvillian superoperator for a set of Lindblad operators.

    A is a vector of traceless, orthogonal Lindblad operators.
    H is an optional Hamiltonian operator.

    Returns the Liouvillian superoperator L corresponding to the
    diagonal-form Lindblad equation

    .. math:: \dot{\rho} = \text{inv\_vec}(L \text{vec}(\rho)) = -i [H, \rho] +\sum_k \left(A_k \rho A_k^\dagger -\frac{1}{2} \{A_k^\dagger A_k, \rho\}\right)
    """
    # James D. Whitfield 2009
    # Ville Bergholm 2009-2010

    # Hamiltonian
    if H == None:
        sh = A[0].shape
        iH = 0
    else:
        sh = H.shape
        iH = 1j * H

    L = zeros(array(sh) ** 2, complex)
    ac = zeros(sh, complex)
    for k in A:
        ac += 0.5 * dot(k.conj().transpose(), k)
        L += lrmul(k, k.conj().transpose()) 

    L += lmul(-ac -iH) +rmul(-ac +iH)
    return L


def superop_fp(L, tol=None):
    r"""Fixed point states of a Liouvillian superoperator.

    Finds the intersection of the kernel of the Liouvillian L
    with the set of valid state operators, thus giving the set of
    fixed point states for the quantum channel represented by the
    master equation

    .. math:: \dot{\rho} = \text{inv\_vec}(L \text{vec}(\rho)).

    Let size(L) == [D, D] (and d = sqrt(D) be the dimension of the Hilbert space).

    Returns the D*n array A, which contains as its columns a set
    of n vectorized orthogonal Hermitian matrices (with respect to
    the Hilbert-Schmidt inner product) which "span" the set of FP
    states in the following sense:

    .. math:: vec(\rho) = A c,  \quad \text{where} \quad c \in \text{R}^n \quad \text{and} \quad c_1 = 1.

    A[:,0] is the shortest vector in the Hermitian kernel of L that
    has trace 1, the other columns of A are traceless and normalized.
    Hence, A defines an (n-1)-dimensional hyperplane.

    A valid state operator also has to fulfill :math:`\rho \ge 0`.
    These operators form a convex set in the Hermitian trace-1 hyperplane
    defined by A. Currently this function does nothing to enforce
    positivity, it is up to the user to choose the coefficients :math:`a_k`
    such that this condition is satisfied.

    Singular values of L less than or equal to the tolerance tol are
    treated as zero.
    """
    #If L has Lindblad form, if L(rho) = \lambda * rho,
    #we also have L(rho') = conj(\lambda) * rho'
    #Hence the non-real eigenvalues of L come in conjugate pairs.
    #Especially if rho \in Ker L, also rho' \in Ker L.

    # Ville Bergholm 2011-2012

    # Hilbert space dimension
    d = sqrt(L.shape[1])

    # Get the kernel of L, restricted to vec-mapped Hermitian matrices.
    # columns of A: complex orthogonal vectors A_i spanning the kernel (with real coefficients)
    A, spectrum = nullspace_hermitian(L, tol)

    # Extract the trace-1 core of A and orthonormalize the rest.
    # We want A_i to be orthonormal wrt. the H-S inner product
    # <X, Y> := trace(X' * Y) = vec(X)' * vec(Y)

    # With this inner product, trace(A) = <eye(d), A>.
    temp = vec(eye(d, d))
    a = dot(temp, A)  # a_i = trace(A_i)

    # Construct the shortest linear combination of A_i which has tr = 1.
    # This is simply (1/d) * eye(d) _iff_ it belongs to span(A_i).
    core = dot(A, (a / norm(a)**2))  # trace-1

    # Remove the component parallel to core from all A_i.
    temp = core / norm(core) # normalized
    A = A -np.outer(temp, dot(temp.conj(), A))

    # Re-orthonormalize the vectors, add the core
    ttt = orth(A, 1e-6)
    A = c_[core, ttt]  # TODO tolerance?

    N = A.shape[1]
    for k in range(N):
        temp = inv_vec(A[:, k])
        A[:, k] = vec(0.5 * (temp + temp.conj().transpose())) # re-Hermitize to fix numerical errors

    # TODO intersect with positive ops

    #[u,t] = schur(L)
    #unnormality = norm(t, 'fro')^2 -norm(diag(t))^2
    #[us, ts] = ordschur(u, t, 'udi')
    #E = ordeig(t)
    # TODO eigendecomposition, find orthogonal complement to span(v) = ker(v').
    # these are the states which do not belong to eigenspaces and
    # should show transient polynomial behavior

    return A, spectrum



# physical operators

@copy_memoize
def angular_momentum(d):
    r"""Angular momentum matrices.

    :param int d: dimension, the corresponding angular momentum quantum number is j = (d-1)/2
    :returns: 3-tuple of angular momentum matrices (Jx, Jy, Jz)

    Returns the dimensionless angular momentum matrices :math:`\vec{J} / \hbar`
    for the d-dimensional subspace.

    The angular momentum matrices fulfill the commutation relation
    :math:`[J_x, J_y] = i J_z` and all its cyclic permutations.
    """
    # Ville Bergholm 2009-2014

    if d < 1:
        raise ValueError('Dimension must be one or greater.')

    j = (d - 1) / 2 # angular momentum quantum number, d == 2*j + 1
    # raising operator in subspace J^2 = j*(j+1)
    m = j
    Jplus = zeros((d, d))
    for k in range(d-1):
        m -= 1
        Jplus[k, k+1] = sqrt(j*(j+1) -m*(m+1))

    # lowering operator
    Jminus = Jplus.conj().transpose()
    # Jplus  = Jx + i*Jy
    # Jminus = Jx - i*Jy
    return (0.5*(Jplus + Jminus), 0.5j*(Jminus - Jplus), diag(np.arange(j, -j-1, -1)))


@copy_memoize
def boson_ladder(d):
    r"""Bosonic ladder operators.

    :param int d: truncation dimension
    :returns: bosonic annihilation operator :math:`b`

    Returns the d-dimensional approximation of the bosonic
    annihilation operator :math:`b` for a single bosonic mode in the
    number basis :math:`\{|0\rangle, |1\rangle, ..., |d-1\rangle\}`.

    The corresponding creation operator is :math:`b^\dagger`.
    The (infinite-dimensional) bosonic annihilation and creation operators fulfill the commutation relation
    :math:`[b, b^\dagger] = I`. Due to the d-dimensional basis truncation,
    this does not hold for the last dimension of the b matrices returned by this function.
    """
    # Ville Bergholm 2009-2010

    return diag(sqrt(range(1, d)), 1)


@copy_memoize
def fermion_ladder(n):
    r"""Fermionic ladder operators.

    :param int n: number of fermionic modes
    :returns: array of fermionic annihilation operators :math:`f_k`

    Returns the fermionic annihilation operators for a
    system of n fermionic modes in the second quantization.

    The annihilation operators are built using the Jordan-Wigner
    transformation for a chain of n qubits, where the state of each
    qubit denotes the occupation number of the corresponding mode.
    First define annihilation and number operators for a lone fermion mode:

    .. math::

      \sigma^- &:= (\sigma_x + i \sigma_y)/2,\\
      n &:= \sigma^+ \sigma^- = (I-\sigma_z)/2,\\
      &\sigma^-|0\rangle = 0, \quad \sigma^-|1\rangle = |0\rangle, \quad n|k\rangle = k|k\rangle.

    Then define a phase operator to keep track of sign changes when
    permuting the order of the operators: :math:`\phi_k := \sum_{j=1}^{k-1} n_j`.
    Now, the fermionic annihilation operators for the n-mode system are given by

    .. math::

      f_k := (-1)^{\phi_k} \sigma^-_k = \left(\bigotimes_{j=1}^{k-1} {\sigma_z}_j \right) \sigma^-_k.

    These operators fulfill the required anticommutation relations:

    .. math::

       \{f_j, f_k\}  &= 0,\\
       \{f_j, f_k^\dagger\} &= I \delta_{jk},\\
       f_k^\dagger f_k &= n_k.
    """
    # Ville Bergholm 2009-2014

    s = array([[0, 1], [0, 0]]) # single annihilation op
    temp = 1

    # empty array for the annihilation operators
    f = empty(n, object)

    # Jordan-Wigner transform
    for k in range(n):
        f[k] = mkron(temp, s, eye(2 ** (n-k-1)))
        temp = kron(temp, sz)

    return f



# SU(2) rotations

def R_nmr(theta, phi):
    r"""SU(2) rotation :math:`\theta_\phi` (NMR notation).

    Returns the one-qubit rotation by angle theta about the unit
    vector :math:`[\cos(\phi), \sin(\phi), 0]`, or :math:`\theta_\phi` in the NMR notation.
    """
    # Ville Bergholm 2009-2014

    return expm(-1j * theta/2 * (np.cos(phi) * sx + np.sin(phi) * sy))


def R_x(theta):
    r"""SU(2) x-rotation.

    Returns the one-qubit rotation about the x axis by the angle theta,
    :math:`e^{-i \sigma_x \theta/2}`.
    """
    # Ville Bergholm 2006-2009

    return expm(-1j * theta/2 * sx)


def R_y(theta):
    r"""SU(2) y-rotation.

    Returns the one-qubit rotation about the y axis by the angle theta,
    :math:`e^{-i \sigma_y \theta/2}`.
    """
    # Ville Bergholm 2006-2009

    return expm(-1j * theta/2 * sy)


def R_z(theta):
    r"""SU(2) z-rotation.

    Returns the one-qubit rotation about the z axis by the angle theta,
    :math:`e^{-i \sigma_z \theta/2}`.
    """
    # Ville Bergholm 2006-2009

    return array([[exp(-1j * theta/2), 0], [0, exp(1j * theta/2)]])



# decompositions

def spectral_decomposition(A):
    r"""Spectral decomposition of a Hermitian matrix.

    Returns the unique eigenvalues a and the corresponding projectors P
    for the Hermitian matrix A, such that :math:`A = \sum_k  a_k P_k`.
    """
    # Ville Bergholm 2010

    d, v = eighsort(A)
    d = d.real  # A is assumed Hermitian

    # combine projectors for degenerate eigenvalues
    a = [d[0]]
    P = [projector(v[:, 0])]
    for k in range(1, len(d)):
        temp = projector(v[:, k])
        if abs(d[k] - d[k-1]) > tol:
            # new eigenvalue, new projector
            a.append(d[k])
            P.append(temp)
        else:
            # same eigenvalue, extend current P
            P[-1] += temp
    return array(a), P



# tensor bases

@copy_memoize
def gellmann(d):
    r"""Gell-Mann matrices.

    Returns the d**2 - 1 (traceless, Hermitian) Gell-Mann matrices of dimension d,
    normalized such that :math:`\mathrm{Tr}(G_i^\dagger G_j) = \delta_{ij}`.
    They form an orthonormal basis for the real vector space of d*d traceless Hermitian matrices.

    The matrices are returned in the array A, arranged such that the n:th Gell-Mann matrix is A[n,:,:].
    """
    # Ville Bergholm 2006-2014

    if d < 1:
        raise ValueError('Dimension must be >= 1.')

    G = zeros((d**2 - 1, d, d), dtype=complex)

    # diagonal
    ddd = zeros(d)
    ddd[0] = 1
    x = 1 / sqrt(2)
    # iterate through the lower triangle
    n = 0
    for k in range(1, d):
        for j in range(0, k):
            # nondiagonal
            G[n, k, j] = x
            G[n, j, k] = x
            n += 1

            G[n, k, j] =  1j * x
            G[n, j, k] = -1j * x
            n += 1
        ddd[k] = -sum(ddd)
        G[n, :, :] = diag(ddd) / norm(ddd)
        ddd[k] = 1
        n += 1

    return G


# TODO lazy evaluation/cache purging would be nice here to control memory usage
tensorbasis_cache = {}
def tensorbasis(n, d=None, get_locality=False):
    r"""Hermitian tensor-product basis for End(H).

    Returns a Hermitian basis for linear operators on the Hilbert space H
    which shares H's tensor product structure. The basis elements are tensor products
    of Gell-Mann matrices (which in the case of qubits are equal to Pauli matrices).
    The basis elements are normalized such that :math:`\mathrm{Tr}(b_i^\dagger b_j) = \delta_{ij}`.

    Input is either two scalars, n and d, in which case the system consists of n qu(d)its, :math:`H = C_d^{\otimes n}`,
    or the vector dim, which contains the dimensions of the individual subsystems:
    :math:`H = C_{dim[0]} \otimes ... \otimes C_{dim[n-1]}`.

    In addition to expanding Hermitian operators on H, this basis can be multiplied by
    the imaginary unit to obtain the antihermitian generators of U(prod(dim)).
    """
    # Ville Bergholm 2005-2011

    if d == None:
        # dim vector
        dim = n
        n = len(dim)
    else:
        # n qu(d)its
        dim = ones(n, int) * d

    # check cache first
    dim = tuple(dim)
    if dim in tensorbasis_cache:
        if get_locality:
            # tuple: (matrices, locality)
            return deepcopy(tensorbasis_cache[dim])
        else:
            # just the matrices
            return deepcopy(tensorbasis_cache[dim][0])

    n_elements = array(dim) ** 2    # number of basis elements for each subsystem, incl. identity
    n_all = prod(n_elements) # number of all tensor basis elements, incl. identity

    B = []
    locality = zeros(n_all, dtype = bool)  # logical array, is the corresponding basis element local?
    # create the tensor basis
    for k in range(n_all):  # loop over all basis elements
        inds = np.unravel_index(k, n_elements)
        temp = 1 # basis element being built
        nonid = 0  # number of non-id. matrices included in this element

        for j in range(n):  # loop over subsystems
            ind = inds[j]   # which local basis element to use
            d = dim[j]

            if ind > 0:
                nonid += 1 # using a non-identity matrix
                L = gellmann(d)[ind - 1]  # Gell-Mann basis vector for the subsystem
                # TODO gellmann copying the entire basis for a single matrix is inefficient...
            else:
                L = eye(d) / sqrt(d)  # identity (scaled)
            temp = kron(temp, L)  # tensor in another matrix

        B.append(temp)
        locality[k] = (nonid < 2) # at least two non-identities => nonlocal element

    # store into cache
    tensorbasis_cache[dim] = deepcopy((B, locality))
    if get_locality:
        return (B, locality)
    else:
        return B



# misc

def op_list(G, dim):
    r"""Operator consisting of k-local terms, given as a list.

    :param list G: list of k-local operator terms
    :param vector dim: vector of subsystem dimensions
    :returns: operator O defined by the connection list

    G is a list of arrays, :math:`G = [c_1, c_2, ..., c_n]`,
    where each array :math:`c_i` corresponds to a term in O.

    An array that has 2 columns and k rows, :math:`c_i` = [(A1, s1), (A2, s2), ... , (Ak, sk)],
    where Aj are arrays and sj subsystem indices, corresponds to the
    k-local term given by the tensor product

    .. math:: A_1^{(s_1)} A_2^{(s_2)} \cdots A_k^{(s_k)}.

    The dimensions of all operators acting on subsystem sj must match dim[sj].

    Alternatively one can think of G as defining a hypergraph, where
    each subsystem corresponds to a vertex and each array c_i in the list
    describes a hyperedge connecting the vertices {s1, s2, ..., sk}.

    Example: The connection list
    G = [[(sz,1)], [(sx,1), (sx,3)], [(sy,1), (sy,3)], [(sz,1), (sz,3)], [(sz,2)], [(A,2), (B+C,3)], [(2*sz,3)]]
    corresponds to the operator

    .. math::

       \sigma_{z1} +\sigma_{z2} +2 \sigma_{z3} +\sigma_{x1} \sigma_{x3} +\sigma_{y1} \sigma_{y3} +\sigma_{z1} \sigma_{z3} +A_2 (B+C)_3.
    """
    # Ville Bergholm 2009-2014

    # TODO we could try to infer dim from the operators
    H = 0j
    for k, spec in enumerate(G):
        a = -1  # last subsystem taken care of
        term = 1
        for j, op in enumerate(spec):
            if len(op) != 2:
                raise ValueError('Malformed local term {0} in spec {1}.'.format(j, k))

            b = op[1]  # subsystem number
            if b <= a:
                raise ValueError('Spec {0} not in ascending order.'.format(k))

            if op[0].shape[1] != dim[b]:
                raise ValueError('The dimension of operator {0} in spec {1} does not match dim.'.format(j, k))

            term = mkron(term, eye(prod(dim[a+1:b])), op[0])
            a = b

        # final identity
        term = mkron(term, eye(prod(dim[a+1:])))
        H += term
    return H


def qubits(n):
    """Dimension vector for an all-qubit system.
    
    For the extemely lazy, returns (2,) * n
    """
    # Ville Bergholm 2010

    return (2,) * n


def majorize(x, y):
    r"""Majorization partial order of real vectors.

    :param vector x, y: real vectors, dimension d
    :returns: True iff :math:`x \preceq y`

    Returns True iff x is majorized by y, denoted by :math:`x \preceq y`. This is equivalent to

    .. math::

       \sum_{k=1}^n x^{\downarrow}_k \le \sum_{k=1}^n y^{\downarrow}_k \quad \text{for all} \quad n \in \{1, 2, \ldots, d\},

    where :math:`x^{\downarrow}` is the vector x with the elements sorted in nonincreasing order.

    :math:`x \preceq y` if and only if x is in the convex hull of all the coordinate permutations of y.
    """
    # Ville Bergholm 2010-2012

    if x.ndim != 1 or y.ndim != 1 or np.iscomplexobj(x) or np.iscomplexobj(y):
        raise ValueError('Inputs must be real vectors.')

    if len(x) != len(y):
        raise ValueError('The vectors must be of equal length.')

    x = np.cumsum(sort(x)[::-1])
    y = np.cumsum(sort(y)[::-1])

    if abs(x[-1] -y[-1]) <= tol:
        # exact majorization
        return all(x-y <= tol)
    else:
        # weak majorization could still be possible, but...
        _warn('Vectors have unequal sums.')
        return False


def mkron(*arg):
    r"""This is how kron should work, dammit.

    Returns the tensor (Kronecker) product :math:`X = A \otimes B \otimes \ldots`
    """
    # Ville Bergholm 2009

    X = 1
    for A in arg:
        X = kron(X, A)
    return X



# -*- coding: utf-8 -*-
"Unit tests for qit.seq"
# Ville Bergholm 2011-2014

import unittest
from numpy import pi
from numpy.random import rand, randn
from numpy.linalg import norm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base import sx, sy, tol
from qit.state import state
from qit.utils import rand_positive
from qit.seq import *



class SeqTest(unittest.TestCase):
    def test_funcs(self):
        """Testing the control sequences module."""

        nmr([[3, 2], [1, 2], [-1, 0.3]])

        # pi rotation
        U = seq2prop(nmr([[pi, 0]]))
        self.assertAlmostEqual(norm(U +1j*sx), 0, delta=tol)
        U = seq2prop(nmr([[pi, pi/2]]))
        self.assertAlmostEqual(norm(U +1j*sy), 0, delta=tol)

        # rotation sequences in the absence of errors
        theta = pi * rand()
        phi = 2*pi * rand()
        U = seq2prop(nmr([[theta, phi]]))
        V = seq2prop(bb1(theta, phi, location = rand()))
        self.assertAlmostEqual(norm(U-V), 0, delta=tol)
        V = seq2prop(corpse(theta, phi))
        self.assertAlmostEqual(norm(U-V), 0, delta=tol)
        V = seq2prop(scrofulous(theta, phi))
        self.assertAlmostEqual(norm(U-V), 0, delta=tol)

        cpmg(2.0, 3)

        # equivalent propagations
        s = state(rand_positive(2))
        seq = scrofulous(pi*rand(), 2*pi*rand())
        s1 = s.u_propagate(seq2prop(seq))
        out, t = propagate(s, seq, base_dt=1)
        s2 = out[-1]
        self.assertAlmostEqual((s1-s2).norm(), 0, delta=tol)



if __name__ == '__main__':
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
r"""
Control sequences (:mod:`qit.seq`)
==================================


Piecewise constant control sequences for quantum systems.
Each control sequence is a dictionary with the following keys:

=======  ===============================================================================================
A        Drift generator (typically :math:`i/\hbar` times a Hamiltonian and a time unit of your choice).
B        List of control generators. c := len(B).
tau      Vector of durations of the time slices. m := len(tau).
control  Array, shape (m, c). control[i,j] is the value of control field j during time slice i.
=======  ===============================================================================================

The total generator for the time slice j is thus given by

.. math::

   G_j = A +\sum_k \text{control}_{jk} B_k,

and the corresponding propagator is

.. math::

   P_j = \exp(-\tau_j G_j).


NOTE the unusual sign convention.


.. currentmodule:: qit.seq

Contents
--------

.. autosummary::

   nmr
   corpse
   bb1
   scrofulous
   cpmg
   seq2prop
   propagate
"""
# Ville Bergholm 2011-2014

from __future__ import division, absolute_import, print_function, unicode_literals

from numpy import array, sin, cos, arcsin, arccos, pi, asarray, eye, zeros, r_, c_, dot, nonzero, ceil, linspace
from scipy.linalg import expm
from scipy.optimize import brentq

from .base import sx, sy, tol


__all__ = ['nmr', 'bb1', 'corpse', 'cpmg', 'scrofulous', 'seq2prop', 'propagate']


def nmr(a):
    r"""Convert NMR-style rotations into a one-qubit control sequence.

    Returns a one-qubit control sequence corresponding to the array a:

    .. math::

       a = [[\theta_1, \phi_1], [\theta_2, \phi_2], ...]

    Each :math:`\theta, \phi` pair corresponds to a NMR rotation
    of the form :math:`\theta_\phi`,
    or a rotation of the angle :math:`\theta`
    about the unit vector :math:`[\cos(\phi), \sin(\phi), 0]`.

    .. math::

       R_{\vec{a}}(\theta) = \exp(-i \vec{a} \cdot \vec{\sigma} \theta/2) = \exp(-i H t) \quad \Leftarrow \quad
       H = \vec{a} \cdot \vec{\sigma}/2, \quad t = \theta.
    """
    # Ville Bergholm 2006-2009

    a = asarray(a, dtype=float)
    theta = a[:, 0]
    phi   = a[:, 1]

    # find theta angles that are negative, convert them to corresponding positive rotations
    rows = nonzero(theta < 0)[0]
    theta[rows] = -theta[rows]
    phi[rows] = phi[rows] + pi

    # construct the sequence TODO make it a class?
    # NOTE the strange sign convention in A and B
    s = {'A': zeros((2, 2)),
         'B': [0.5j * sx, 0.5j * sy],
         'tau': theta,
         'control': c_[cos(phi), sin(phi)]
         }
    return s 


def bb1(theta, phi=0, location=0.5):
    r"""Sequence for correcting pulse length errors.

    Returns the Broadband number 1 control sequence for correcting
    proportional errors in pulse length (or amplitude) :cite:`Wimperis`.

    The target rotation is :math:`\theta_\phi` in the NMR notation.
    """
    # Ville Bergholm 2009-2012

    ph1 = arccos(-theta / (4*pi))
    W1  = [[pi, ph1], [2*pi, 3*ph1], [pi, ph1]]
    return nmr([[location * theta, phi]] + W1 + [[(1-location) * theta, phi]])


def corpse(theta, phi=0):
    r"""Sequence for correcting off-resonance errors.

    Returns the CORPSE control sequence for correcting off-resonance
    errors, i.e. ones arising from a constant but unknown
    :math:`\sigma_z` bias in the Hamiltonian :cite:`Cummins`

    The target rotation is :math:`\theta_\phi` in the NMR notation.

    CORPSE: Compensation for Off-Resonance with a Pulse SEquence
    """
    # Ville Bergholm 2009

    n = [1, 1, 0] # CORPSE
    #n = [0, 1, 0] # short CORPSE

    temp = arcsin(sin(theta / 2) / 2)

    th1 = 2*pi*n[0] +theta/2 -temp
    th2 = 2*pi*n[1] -2*temp
    th3 = 2*pi*n[2] +theta/2 -temp
    return nmr([[th1, phi], [th2, phi+pi], [th3, phi]])


def cpmg(t, n):
    r"""Carr-Purcell-Meiboom-Gill sequence.

    Returns the Carr-Purcell-Meiboom-Gill sequence of n repeats with waiting time t.
    The purpose of the CPMG sequence is to facilitate a T_2 measurement
    under a nonuniform z drift, it is not meant to be a full memory protocol.
    The target operation for this sequence is identity.
    """
    # Ville Bergholm 2007-2012

    s = nmr([[pi/2, pi/2]]) # initial y rotation

    # step: wait, pi x rotation, wait
    step_tau  = array([t, pi, t])
    step_ctrl = array([[0, 0], [1, 0], [0, 0]])
    for k in range(n):
        s['tau'] = r_[s['tau'], step_tau]
        s['control'] = r_[s['control'], step_ctrl]
    return s


def scrofulous(theta, phi=0):
    r"""Sequence for correcting pulse length errors.

    Returns the SCROFULOUS control sequence for correcting errors
    in pulse duration (or amplitude) :cite:`Cummins`.

    The target rotation is :math:`\theta_\phi` in the NMR notation.

    SCROFULOUS: Short Composite ROtation For Undoing Length Over- and UnderShoot
    """
    # Ville Bergholm 2006-2014

    th1 = brentq(lambda t: (sin(t)/t -(2 / pi) * cos(theta / 2)), 0.1, 4.6)
    ph1 = arccos(-pi * cos(th1) / (2 * th1 * sin(theta / 2)))
    ph2 = ph1 - arccos(-pi / (2 * th1))

    u1 = [[th1, ph1 +phi]]
    u2 = [[pi,  ph2 +phi]]
    return nmr(u1 + u2 + u1)


def seq2prop(s):
    r"""Propagator corresponding to a control sequence.

    Returns the propagator matrix corresponding to the
    action of the control sequence s.

    Governing equation: :math:`\dot(X)(t) = -(A +\sum_k u_k(t) B_k) X(t) = -G(t) X(t)`.
    """
    # Ville Bergholm 2009-2012

    A = s['A']
    B = s['B']

    n = len(s['tau'])
    P = eye(A.shape[0])
    for j in range(n):
        G = A
        for k, b in enumerate(B):
            G = G + s['control'][j, k] * b

        temp = expm(-s['tau'][j] * G)  # NOTE the sign convention here
        P = dot(temp, P)

    return P


def propagate(s, seq, out_func=lambda x: x, base_dt=0.1):
    """Propagate a state in time using a control sequence.
    
    If no output function is given, we use an identity map.
    """
    # Ville Bergholm 2009-2014

    A = seq['A']
    B = seq['B']

    n = len(seq['tau'])
    t = [0]  # initial time
    out = [out_func(s)]  # initial state

    # loop over the sequence
    for j in range(n):
        G = A
        for k, b in enumerate(B):
            G = G + seq['control'][j, k] * b

        T = seq['tau'][j]  # pulse duration
        n_steps = max(int(ceil(T / base_dt)), 1)
        dt = T / n_steps

        P = expm(-G * dt)  # NOTE TODO the sign convention here
        for k in range(n_steps):
            s = s.u_propagate(P)
            out.append(out_func(s))

        temp = t[-1]
        t.extend(list(linspace(temp+dt, temp+T, n_steps)))
    return out, t



#! /usr/bin/python
from setuptools import setup
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))

# Read the version number from a source file.
def find_version(*file_paths):
    # Use codecs.open for Python 2 compatibility
    with codecs.open(os.path.join(here, *file_paths), 'r', 'utf_8') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Use README.txt as the long description
with codecs.open('README.txt', 'r', 'utf_8') as f:
    long_description = f.read()    


setup(name             = 'qit',
      version          = find_version('qit', '__init__.py'),
      description      = 'Quantum Information Toolkit',
      long_description = long_description,
      url              = 'http://qit.sourceforge.net/',
      author           = 'Ville Bergholm et al.',
      author_email     = 'smite-meister@users.sourceforge.net',
      classifiers      = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Information Technology',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        ],
      keywords = 'quantum information, quantum mechanics, toolkit',
      packages         = ['qit'],
      install_requires = ['numpy>=1.7.1', 'scipy>=0.11', 'matplotlib>=1.2']
      )



# -*- coding: utf-8 -*-
"Unit tests for qit.gate"
# Ville Bergholm 2010-2014

import unittest
from numpy import prod
from numpy.random import rand, randn
from numpy.linalg import norm

# HACK to import the module in this source distribution, not the installed one!
import sys, os
sys.path.insert(0, os.path.abspath('.'))

from qit import version
from qit.base import sx, sy, sz, tol
from qit.gate import *


class GateTest(unittest.TestCase):
    def test_funcs(self):
        """Testing the quantum gates module."""

        dim = (2, 4)
        D = prod(dim)

        I = id(dim)
        S = swap(*dim)
        # swap' * swap = I
        self.assertAlmostEqual((S.ctranspose() * S -I).norm(), 0, delta=tol)

        # TODO test the output
        U = phase(randn(prod(dim)), dim)
        U = walsh(3)
        U = qft(dim)
        U = mod_inc(3, dim, 5)
        V = mod_mul(2, dim, 5)
        U = mod_add(2, 4, 3)
        x = dist(U, V)
        U = controlled(sz, (1, 0), dim)
        cnot = controlled(sx)
        U = single(sy, 0, dim)
        U = two(cnot, (2, 0), (2, 3, 2))

        # test bad input
        self.assertRaises(ValueError, dist, I, S)  # output dimension mismatch
        self.assertRaises(ValueError, phase, rand(D-1), dim)  # dimension mismatch
        self.assertRaises(ValueError, mod_inc, 1, 3, 4)  # N too large
        self.assertRaises(ValueError, mod_mul, 2, 4, 5)  # N too large
        self.assertRaises(ValueError, mod_mul, 2, 4)     # a and N not coprime
        self.assertRaises(ValueError, mod_add, 2, 3, 4)  # N too large
        self.assertRaises(ValueError, controlled, U, (0,), dim)    # ctrl shorter than dim
        self.assertRaises(ValueError, controlled, U, (0, 4), dim)  # ctrl on nonexistant state
        self.assertRaises(ValueError, single, sx, 1, dim)  # input dimension mismatch
        self.assertRaises(ValueError, two, S, (0, 1), (2, 3, 4))   # input dimension mismatch
        self.assertRaises(ValueError, two, S, (-1, 2), (2, 3, 4))  # bad targets
        self.assertRaises(ValueError, two, S, (3, 2), (2, 3, 4))   # bad targets
        self.assertRaises(ValueError, two, S, (0,), (2, 3, 4))     # wrong number of targets



if __name__ == '__main__':
    print('Testing QIT version ' + version())
    unittest.main()



# -*- coding: utf-8 -*-
"""
Local invariants (:mod:`qit.invariant`)
=======================================

This module contains tools for computing and plotting the values of
various local gate and state invariants.

.. currentmodule:: qit.invariant

Contents
--------

.. autosummary::

   LU
   canonical
   makhlin
   max_concurrence
   plot_makhlin_2q
   plot_weyl_2q   
"""
# Ville Bergholm 2011

from __future__ import division, absolute_import, print_function, unicode_literals

import numpy as np
from numpy import (array, asarray, arange, empty, zeros, ones, sqrt, sin, cos, dot, sort, trace, kron,
    pi, r_, c_, linspace, meshgrid, roll, concatenate, angle)
from numpy.linalg import det, eigvals
from scipy.linalg import norm
import matplotlib.pyplot as plt

from .base import sy, Q_Bell
from .lmap import lmap, tensor


# TODO these function names are terrible
__all__ = ['canonical', 'makhlin', 'max_concurrence', 'plot_weyl_2q', 'plot_makhlin_2q', 'LU']


def LU(rho, k, perms):
    r"""Local unitary polynomial invariants of quantum states.

    Computes the permutation invariant :math:`I_{k; \pi_1, \pi_2, \ldots, \pi_n}` for the state :math:`\rho`.
    perms is a tuple containing n k-permutation tuples.

    Example: :math:`I_{3; (123),(12)}(\rho)` = LU_inv(rho, 3, [(1, 2, 0), (1, 0, 2)])

    This function can be very inefficient for some invariants, since
    it does no partial traces etc. which might simplify the calculation.

    Uses the algorithm in :cite:`BBL2012`.
    """
    # Ville Bergholm 2011-2012

    def tensor_pow(rho, n):
        """Returns $\rho^{\otimes n}$."""
        rho.to_op(inplace=True)
        ret = lmap(rho)
        for _ in range(1, n):
            ret = tensor(ret, rho)
        return ret


    n = len(perms)
    if n != rho.subsystems():
        raise ValueError('Need one permutation per subsystem.')

    # convert () to identity permutation
    id_perm = tuple(range(k))
    perms = list(perms)
    for j, p in enumerate(perms):
        if len(p) == 0:
            perms[j] = id_perm

    # splice k sequential copies of the entire system into k copies of each subsystem
    s = arange(n * k).reshape((k, n)).flatten()

    # permute the k copies of each subsystem
    temp = kron(k * arange(n), ones(k, int))
    p = asarray(perms).flatten() + temp

    # Permutations: a*b = a(b), x = y * z^{-1}  <=>  x * z = x(z) = y.
    s_inv = empty(s.shape, int)
    s_inv[s] = arange(n*k)
    total = s_inv[p[s]] # total = s^{-1} * p * s

    # TODO this could be done much more efficiently
    return tensor_pow(rho, k).reorder((total, None)).trace()



def canonical(U):
    """Canonical local invariants of a two-qubit gate.

    Returns a vector of three real canonical local invariants for the
    U(4) matrix U, normalized to the range [0,1].
    Uses the algorithm in :cite:`Childs`.
    """
    # Ville Bergholm 2004-2010

    sigma = kron(sy, sy)
    U_flip = dot(dot(sigma, U.transpose()), sigma)  # spin flipped U
    temp = dot(U, U_flip) / sqrt(complex(det(U)))

    Lambda = eigvals(temp) #[exp(i*2*phi_1), etc]
    # logarithm to the branch (-1/2, 3/2]
    Lambda = angle(Lambda) / pi # divide pi away
    for k in range(len(Lambda)):
        if Lambda[k] <= -0.5:
            Lambda[k] += 2
    S = Lambda / 2
    S = sort(S)[::-1]  # descending order

    n = int(round(sum(S)))  # sum(S) must be an integer
    # take away extra translations-by-pi
    S -= r_[ones(n), zeros(4-n)]
    # put the elements in the correct order
    S = roll(S, -n)

    M = [[1, 1, 0], [1, 0, 1], [0, 1, 1]] # scaled by factor 2
    c = dot(M, S[:3])
    # now 0.5 >= c[0] >= c[1] >= |c[2]|
    # and into the Berkeley chamber using a translation and two Weyl reflections
    if c[2] < 0:
        c[0] = 1 - c[0]
        c[2] = -c[2]
    return c


def makhlin(U):
    """Makhlin local invariants of a two-qubit gate.

    Returns a vector of the three real Makhlin invariants (see :cite:`Makhlin`) corresponding
    to the U(4) gate U.

    Alternatively, given a vector of canonical invariants normalized to [0, 1],
    returns the corresponding Makhlin invariants (see :cite:`Zhang`).
    """
    # Ville Bergholm 2004-2010

    if U.shape[-1] == 3:
        c = U
        # array consisting of vectors of canonical invariants
        c *= pi
        g = empty(c.shape)
        
        g[..., 0] = (cos(c[..., 0]) * cos(c[..., 1]) * cos(c[..., 2])) ** 2 -(sin(c[..., 0]) * sin(c[..., 1]) * sin(c[..., 2])) ** 2
        g[..., 1] = 0.25 * sin(2 * c[..., 0]) * sin(2 * c[..., 1]) * sin(2 * c[..., 2])
        g[..., 2] = 4 * g[..., 0] - cos(2 * c[..., 0]) * cos(2 * c[..., 1]) * cos(2*c[..., 2])
    else:
        # U(4) gate matrix    
        V = dot(Q_Bell.conj().transpose(), dot(U, Q_Bell))
        M = dot(V.transpose(), V)

        t1 = trace(M) ** 2
        t2 = t1 / (16 * det(U))
        g = array([t2.real, t2.imag, ((t1 - trace(dot(M, M))) / (4 * det(U))).real])
    return g


def max_concurrence(U):
    """Maximum concurrence generated by a two-qubit gate.

    Returns the maximum concurrence generated by the two-qubit
    gate U (see :cite:`Kraus`), starting from a tensor state.

    Alternatively, U may be given in terms of a vector of three
    canonical local invariants.
    """
    # Ville Bergholm 2006-2010

    if U.shape[-1] == 4:
        # gate into corresponding invariants
        c = canonical(U)
    else:
        c = U
    temp = roll(c, 1, axis = -1)
    return np.max(abs(sin(pi * concatenate((c - temp, c + temp), axis = -1))), axis = -1)


def plot_makhlin_2q(sdiv=31, tdiv=31):
    """Plots the set of two-qubit gates in the space of Makhlin invariants.

    Plots the set of two-qubit gates in the space of Makhlin
    invariants (see :func:`makhlin`), returns the Axes3D object.

    The input parameters are the s and t divisions of the mesh.
    """
    # Ville Bergholm 2006-2011

    import matplotlib.cm as cm
    import matplotlib.colors as colors

    s = linspace(0, pi,   sdiv)
    t = linspace(0, pi/2, tdiv)

    # more efficient than meshgrid
    #g1 = kron(cos(s).^2, cos(t).^4) - kron(sin(s).^2, sin(t).^4)
    #g2 = 0.25*kron(sin(2*s), sin(2*t).^2)
    #g3 = 4*g1 - kron(cos(2*s), cos(2*t).^2)
    #S = kron(s, ones(size(t)))
    #T = kron(ones(size(s)), t)

    # canonical coordinate plane (s, t, t) gives the entire surface of the set of gate equivalence classes
    S, T = meshgrid(s, t)
    c = c_[S.ravel(), T.ravel(), T.ravel()]
    G = makhlin(c).reshape(sdiv, tdiv, 3)
    C = max_concurrence(c).reshape(sdiv, tdiv)

    fig = plt.gcf()
    ax = fig.add_subplot(111, projection='3d')

    # mesh, waterfall?
    polyc = ax.plot_surface(G[:, :, 0], G[:, :, 1], G[:, :, 2], rstride = 1, cstride = 1,
        cmap = cm.jet, norm = colors.Normalize(vmin=0, vmax=1, clip=True), alpha = 0.6)
    polyc.set_array(C.ravel() ** 2)  # FIXME colors
    ax.axis('equal')
    #ax.axis([-1, 1, -0.5, 0.5, -3, 3])
    #ax.shading('interp')

    ax.set_xlabel('$g_1$')
    ax.set_ylabel('$g_2$')
    ax.set_zlabel('$g_3$')
    plt.title('Makhlin stingray')

    # labels
    ax.text(1.05, 0, 2.7, 'I')
    ax.text(-1.05, 0, -2.7, 'SWAP')
    ax.text(-0.1, 0, 1.2, 'CNOT')
    ax.text(0.1, 0, -1.2, 'DCNOT')
    ax.text(0.1, 0.26, 0, 'SWAP$^{1/2}$')
    ax.text(0, -0.26, 0, 'SWAP$^{-1/2}$')

    fig.colorbar(polyc, ax = ax)
    plt.show()
    return ax


def plot_weyl_2q(ax=None):
    """Plots the two-qubit Weyl chamber.

    Plots the Weyl chamber for the local invariants
    of 2q gates. See :cite:`Zhang`.

    Returns the Axes3D object.
    """
    # Ville Bergholm 2005-2012

    if ax == None:
        ax = plt.subplot(111, projection='3d')
    ax.hold(True)
    ax.plot_surface(array([[0, 0.5, 1], [0, 0.5, 1]]), array([[0, 0, 0], [0, 0.5, 0]]), array([[0, 0, 0], [0, 0.5, 0]]), alpha = 0.2)
    ax.plot_surface(array([[0, 0.5], [0, 0.5]]), array([[0, 0.5], [0, 0.5]]), array([[0, 0], [0, 0.5]]), alpha = 0.2)
    ax.plot_surface(array([[0.5, 1], [0.5, 1]]), array([[0.5, 0], [0.5, 0]]), array([[0, 0], [0.5, 0]]), alpha = 0.2)
    #axis([0 1 0 0.5 0 0.5])
    ax.axis('equal')
    ax.set_xlabel('$c_1/\\pi$')
    ax.set_ylabel('$c_2/\\pi$')
    ax.set_zlabel('$c_3/\\pi$')
    plt.title('Two-qubit Weyl chamber')

    ax.text(-0.05, -0.05, 0, 'I')
    ax.text(1.05, -0.05, 0, 'I')
    ax.text(0.45, 0.55, 0.55, 'SWAP')
    ax.text(0.45, -0.05, 0, 'CNOT')
    ax.text(0.45, 0.55, -0.05, 'DCNOT')
    ax.text(0.20, 0.25, 0, 'SWAP$^{1/2}$')
    ax.text(0.75, 0.25, 0, 'SWAP$^{-1/2}$')
    return ax
